# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file provides an API enabling the use of classical shadows. The original
idea is described in H.Y. Huang, R. Kueng, and J. Preskill, Nature Physics 16,
1050 (2020).
"""

import abc
import numpy as np

from tangelo.linq.circuit import Circuit


class ClassicalShadow(abc.ABC):
    """Abstract class for the classical shadows implementation. Classical
    shadows is a mean to characterize a quantum state (within an error treshold)
    with the fewest measurement possible.
    """

    def __init__(self, circuit=None, bitstrings=None, unitaries=None):
        """Default constructor for the ClassicalShadow object. This class is
        the parent class for the different classical shadows flavors. The object
        is defined by the bistrings and unitaries used in the process. Abstract
        methods are defined to take into account the procedure to inverse the
        channel.

        Args:
            circuit (Circuit): State to characterize.
            bistrings (list of str): Representation of the outcomes for all
                snapshots. E.g. ["11011", "10000", ...].
            unitaries (list of str): Representation of the unitary for every
                snapshot, used to reverse the channel.
        """

        self.circuit = circuit
        self.bitstrings = list() if bitstrings is None else bitstrings
        self.unitaries = list() if unitaries is None else unitaries

        assert len(self.bitstrings) == len(self.unitaries), f"bistrings and unitaries must be the same length."

        # If the state has been estimated, it is stored into this attribute.
        self.state_estimate = None

    @property
    def n_qubits(self):
        """Returns the number of qubits the shadow represents."""
        return self.circuit.width if self.circuit else len(self.bitstrings[0])

    @property
    def size(self):
        """Number of shots used to make the shadow."""
        return len(self.bitstrings)

    @property
    def unique_unitaries(self):
        """Returns the list of unique unitaries."""
        return list(set(self.unitaries))

    def __len__(self):
        """Same as the shadow size."""
        return self.size

    def append(self, bitstring, unitary):
        """Append method to merge new snapshots to an existing shadow.

        Args:
            bistring (str or list of str): Representation of outcomes.
            unitary (str or list of str): Relevant unitary for those outcomes.
        """
        if isinstance(bitstring, list) and isinstance(unitary, list):
            assert len(bitstring) == len(unitary)
            self.bitstrings += bitstring
            self.unitaries += unitary
        elif isinstance(bitstring, str) and isinstance(unitary, str):
            self.bitstrings.append(bitstring)
            self.unitaries.append(unitary)
        else:
            raise ValueError("bistring and unitary arguments must be consistent strings or list of strings.")

    def get_observable(self, qubit_op, *args, **kwargs):
        """Getting an estimated observable value for a qubit operator from the
        classical shadow. This function loops through all terms and calls, for
        each of them, the get_term_observable method defined in the child class.
        Other arguments (args, kwargs) can be passed to the method.

        Args:
            qubit_op (QubitOperator): Operator to estimate.
        """
        observable = 0.
        for term, coeff in qubit_op.terms.items():
            observable += self.get_term_observable(term, coeff, *args, **kwargs)

        return observable

    def simulate(self, backend, initial_statevector=None):
        """Simulate, using a predefined backend, a shadow from a circuit or a
        statevector.

        Args:
            backend (Simulator): Backend for the simulation of a shadow.
            initial_statevector(list/array) : A valid statevector in the format
                supported by the target backend.
        """

        if not self.unitaries:
            raise ValueError(f"The build method of {self.__class__.__name__} must be called before simulation.")

        if self.bitstrings:
            raise NotImplementedError("Appending new simulation results to an already defined self.bistrings is not implemented yet.")

        self.bitstrings = np.chararray(len(self.unitaries), itemsize=self.n_qubits)

        # Different behavior if circuit or initial_statevector is defined.
        circuit_template = self.circuit if self.circuit is not None else Circuit(n_qubits=self.n_qubits)

        for basis_circuit, pauli_word, n_repeat in self.get_basis_circuits(only_unique=True):
            circuit = circuit_template + basis_circuit if (basis_circuit.size > 0) else circuit_template

            backend.n_shots = n_repeat
            freqs, _ = backend.simulate(circuit, initial_statevector=initial_statevector)

            new_bitstrings = list()
            for k, v in freqs.items():
                new_bitstrings += round(v*n_repeat)*[k]
            assert len(new_bitstrings) == n_repeat
            np.random.shuffle(new_bitstrings)

            indices = np.where(np.array(self.unitaries) == pauli_word)[0]
            # Guard rail to ensure that all the unitaries are filled out.
            assert len(indices) == n_repeat

            self.bitstrings[indices] = new_bitstrings

        self.bitstrings = list(self.bitstrings.astype(str))

    @abc.abstractmethod
    def build(self):
        pass

    @abc.abstractmethod
    def get_basis_circuits(self, only_unique=False):
        pass

    @abc.abstractmethod
    def get_term_observable(self):
        pass



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from tangelo.algorithms.variational import SA_OO_Solver, BuiltInAnsatze
from tangelo.toolboxes.molecular_computation.molecule import SecondQuantizedMolecule

Li2 = [('Li', (0, 0, 0)), ('Li', (3.5, 0, 0))]
mol = SecondQuantizedMolecule(Li2, q=0, spin=0, basis="6-31g(d,p)", frozen_orbitals=[0, 1]+list(range(4, 28)))


class SA_OO_SolverTest(unittest.TestCase):

    def test_build(self):
        """Try instantiating SA_OO_Solver with basic input."""

        opt_dict = {"molecule": mol, "ref_states": [[1, 1, 0, 0], [1, 0, 1, 0]], "max_cycles": 5}
        sa_oo_solver = SA_OO_Solver(opt_dict)
        sa_oo_solver.build()

        # Test error is raised when no molecule is provided
        opt_dict = {"max_cycles": 15}
        self.assertRaises(ValueError, SA_OO_Solver, opt_dict)

    def test_li2(self):
        """Try reproducing Li2 2 electrons in 2 orbitals casscf for both states from pyscf
        """

        opt_dict = {"molecule": mol, "ref_states": [[1, 1, 0, 0], [1, 0, 1, 0]], "qubit_mapping": "bk", "up_then_down": False,
                    "tol": 1.e-5, "ansatz": BuiltInAnsatze.UCCGD, "weights": [1, 1], "n_oo_per_iter": 3}
        sa_oo_solver = SA_OO_Solver(opt_dict)
        sa_oo_solver.build()
        sa_oo_solver.iterate()

        # Code to generate exact pyscf results
        # mol = pyscf.M(atom=Li2, basis='6-31g(d,p)', spin=0)
        # myhf=mol.RHF().run()
        # weights = np.ones((2))/2
        # ncas, nelecas = (2, (1,1))
        # mycas = myhf.CASSCF(ncas, nelecas).state_average_(weights)
        # res = mycas.kernel()
        # mc = myhf.CASCI(ncas, nelecas)
        # mc.fcisolver.nroots = 2
        # exact_energies = mc.casci(mycas.mo_coeff)[0]
        exact_energies = [-14.87324, -14.85734]

        self.assertAlmostEqual(sa_oo_solver.state_energies[0], exact_energies[0], places=4)
        self.assertAlmostEqual(sa_oo_solver.state_energies[1], exact_energies[1], places=4)

        oo_resources = sa_oo_solver.get_resources()
        self.assertEqual(oo_resources["circuit_width"], 4)
        self.assertEqual(oo_resources["vqe_variational_parameters"], 3)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import os

import numpy as np
from openfermion import load_operator

from tangelo.molecule_library import mol_H2_sto3g
from tangelo.toolboxes.operators import QubitOperator
from tangelo.toolboxes.qubit_mappings import jordan_wigner, symmetry_conserving_bravyi_kitaev
from tangelo.toolboxes.ansatz_generator import VSQS
from tangelo.linq import Simulator
from tangelo.toolboxes.qubit_mappings.statevector_mapping import get_reference_circuit

# For openfermion.load_operator function.
pwd_this_test = os.path.dirname(os.path.abspath(__file__))


class VSQSTest(unittest.TestCase):

    def test_vsqs_set_var_params(self):
        """Verify behavior of set_var_params for different inputs (list, numpy array).
        """

        vsqs_ansatz = VSQS(mol_H2_sto3g)

        two_ones = np.ones((2,))

        vsqs_ansatz.set_var_params([1., 1.])
        np.testing.assert_array_almost_equal(vsqs_ansatz.var_params, two_ones, decimal=6)

        vsqs_ansatz.set_var_params(np.array([1., 1.]))
        np.testing.assert_array_almost_equal(vsqs_ansatz.var_params, two_ones, decimal=6)

    def test_vsqs_incorrect_number_var_params(self):
        """Return an error if user provide incorrect number of variational
        parameters.
        """

        vsqs_ansatz = VSQS(mol_H2_sto3g)

        self.assertRaises(ValueError, vsqs_ansatz.set_var_params, np.array([1., 1., 1., 1.]))

    def test_vsqs_H2(self):
        """Verify closed-shell VSQS functionalities for H2."""

        # Build circuit
        vsqs_ansatz = VSQS(mol_H2_sto3g, intervals=3, time=3)
        vsqs_ansatz.build_circuit()

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = jordan_wigner(mol_H2_sto3g.fermionic_hamiltonian)

        # Assert energy returned is as expected for given parameters
        sim = Simulator()
        vsqs_ansatz.update_var_params([0.66666667, 0.9698286, 0.21132472, 0.6465473])
        energy = sim.get_expectation_value(qubit_hamiltonian, vsqs_ansatz.circuit)
        self.assertAlmostEqual(energy, -1.1372701255155757, delta=1e-6)

    def test_vsqs_H4_doublecation(self):
        """Verify closed-shell VSQS functionalities for H4 2+ by using saved qubit hamiltonian and initial hamiltonian"""

        var_params = [-2.53957674, 0.72683888, 1.08799500, 0.49836183,
                      -0.23020698, 0.93278630, 0.50591026, 0.50486903]

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = load_operator("mol_H4_doublecation_minao_qubitham_jw_b.data", data_directory=pwd_this_test+"/data", plain_text=True)
        initial_hamiltonian = load_operator("mol_H4_doublecation_minao_init_qubitham_jw_b.data", data_directory=pwd_this_test+"/data", plain_text=True)
        reference_state = get_reference_circuit(8, 2, "jw", up_then_down=True, spin=0)

        # Build circuit
        vsqs_ansatz = VSQS(qubit_hamiltonian=qubit_hamiltonian, h_init=initial_hamiltonian, reference_state=reference_state,
                           intervals=5, time=5, trotter_order=2)
        vsqs_ansatz.build_circuit()

        # Assert energy returned is as expected for given parameters
        sim = Simulator()
        vsqs_ansatz.update_var_params(var_params)
        energy = sim.get_expectation_value(qubit_hamiltonian, vsqs_ansatz.circuit)
        self.assertAlmostEqual(energy, -0.85425, delta=1e-4)

    def test_vsqs_H2_with_h_nav(self):
        """Verify closed-shell VSQS functionalities for H2 with navigator hamiltonian"""
        navigator_hamiltonian = (QubitOperator('X0 Y1', 0.03632537110234512) + QubitOperator('Y0 X1', 0.03632537110234512)
                              + QubitOperator('Y0', 2.e-5) + QubitOperator('Y1', 2.e-5))

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = symmetry_conserving_bravyi_kitaev(mol_H2_sto3g.fermionic_hamiltonian, 4, 2, False, 0)

        # Build circuit
        vsqs_ansatz = VSQS(mol_H2_sto3g, intervals=2, time=1, mapping='scbk', up_then_down=True, trotter_order=2,
                           h_nav=navigator_hamiltonian)
        vsqs_ansatz.build_circuit()

        # Assert energy returned is as expected for given parameters
        sim = Simulator()
        vsqs_ansatz.update_var_params([0.50000001, -0.02494214, 3.15398767])
        energy = sim.get_expectation_value(qubit_hamiltonian, vsqs_ansatz.circuit)
        self.assertAlmostEqual(energy, -1.1372701255155757, delta=1e-6)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


def rotosolve_step(func, var_params, i, *func_args):
    """Gradient free optimization step using specific points to
    characterize objective function w.r.t to parameter values. Based on
    formulas in arXiv:1905.09692, Mateusz Ostaszewski

        Args:
            func (function handle): The function that performs energy
                estimation. This function takes variational params as input
                and returns a float.
            var_params (list of float): The variational parameters.
            i (int): Index of the variational parameter to update.
            *func_args (tuple): Optional arguments to pass to func.
        Returns:
            list of floats: Optimal parameters.
    """

    # Charaterize sinusoid of objective function using specific parameters
    var_params[i] = 0
    m_1 = func(var_params, *func_args)

    var_params[i] = 0.5 * np.pi
    m_2 = func(var_params, *func_args)

    var_params[i] = -0.5 * np.pi
    m_3 = func(var_params, *func_args)

    # Calculate theta_min based on measured values
    theta_min = -0.5 * np.pi - np.arctan2(2. * m_1 - m_2 - m_3, m_2 - m_3)

    if theta_min < -np.pi:
        theta_min += 2 * np.pi
    elif theta_min > np.pi:
        theta_min -= 2 * np.pi

    # Update parameter to theta_min
    var_params[i] = theta_min

    return var_params


def rotosolve(func, var_params, *func_args, ftol=1e-5, maxiter=100):
    """Optimization procedure for parameterized quantum circuits whose
     objective function varies sinusoidally with the parameters. Based
     on the work by arXiv:1905.09692, Mateusz Ostaszewski.

    Args:
        func (function handle): The function that performs energy
            estimation. This function takes variational parameters as input
            and returns a float.
        var_params (list): The variational parameters.
        ftol (float): Convergence threshold.
        maxiter (int): The maximum number of iterations.
        *func_args (tuple): Optional arguments to pass to func.

    Returns:
        float: The optimal energy found by the optimizer.
        list of floats: Optimal parameters.
     """
    # Get intial value, and run rotosolve for up to maxiter iterations
    energy_old = func(var_params, *func_args)
    for it in range(maxiter):
        # Update parameters one at a time using rotosolve_step
        for i in range(len(var_params)):
            var_params = rotosolve_step(func, var_params, i, *func_args)
        energy_new = func(var_params, *func_args)

        # Check if convergence tolerance is met
        if abs(energy_new - energy_old) <= ftol:
            break

        # Update energy value
        energy_old = energy_new

    return energy_new, var_params



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that defines the SA-OO-VQE algorithm

Ref:
[1] Saad Yalouz, Bruno Senjean, Jakob Gunther, Francesco Buda, Thomas E. O'Brien, Lucas Visscher, "A state-averaged
orbital-optimized hybrid quantum-classical algorithm for a democratic description of ground and excited states",
2021, Quantum Sci. Technol. 6 024004

"""

from itertools import product

import numpy as np
from scipy.linalg import expm

from tangelo.algorithms.variational import SA_VQESolver


class SA_OO_Solver(SA_VQESolver):
    """State Averaged Orbital Optimized Solver class. This is an iterative algorithm that uses SA-VQE alternatively with an
    orbital optimization step.

    Users must first set the desired options of the SA_OO_Solver object through the
    __init__ method, and call the "build" method to build the underlying objects
    (mean-field, hardware backend, ansatz...). They are then able to call any of
    the energy_estimation, simulate, get_rdm, or iterate methods. In particular, iterate
    runs the SA-OO algorithm, alternating calls to the SA_VQESolver and orbital optimization.

    Attributes:
        tol (float): Maximum energy difference before convergence
        max_cycles (int): Maximum number of iterations for sa-oo-vqe
        n_oo_per_iter (int): Number of orbital optimization Newton-Raphson steps per SA-OO-VQE iteration
        molecule (SecondQuantizedMolecule) : the molecular system.
        qubit_mapping (str) : one of the supported qubit mapping identifiers.
        ansatz (Ansatze) : one of the supported ansatze.
        optimizer (function handle): a function defining the classical optimizer and its behavior.
        initial_var_params (str or array-like) : initial value for the classical optimizer.
        backend_options (dict) : parameters to build the tangelo.linq Simulator class.
        penalty_terms (dict): parameters for penalty terms to append to target qubit Hamiltonian (see penalty_terms
            for more details).
        ansatz_options (dict): parameters for the given ansatz (see given ansatz file for details).
        up_then_down (bool): change basis ordering putting all spin up orbitals first, followed by all spin down.
            Default, False has alternating spin up/down ordering.
        qubit_hamiltonian (QubitOperator-like): Self-explanatory.
        verbose (bool): Flag for VQE verbosity.
        ref_states (list): The vector occupations of the reference configurations
        weights (array): The weights of the occupations
     """

    def __init__(self, opt_dict: dict):

        oo_options = {"tol": 1e-3,
                      "max_cycles": 15,
                      "n_oo_per_iter": 1}

        if "molecule" not in opt_dict:
            raise ValueError(f"A molecule must be provided for {self.__class__.__name__}")

        # remove SA-OO-VQE specific options before calling SA_VQESolver.__init__() and move values to oo_options
        opt_dict_sa_vqe = opt_dict.copy()
        for k, v in opt_dict.items():
            if k in oo_options:
                oo_options[k] = opt_dict_sa_vqe.pop(k)

        # Initialization of SA_VQESOLVER will check if spurious dictionary items are present
        super().__init__(opt_dict_sa_vqe)

        # Add oo_options to attributes
        for k, v in oo_options.items():
            setattr(self, k, v)

        self.n_ref_states = len(self.ref_states)

        self.converged = False
        self.iteration = 0
        self.energies = list()
        # vqe_energies could include a penalty term contribution so will be different from energies calculated using rdms
        self.vqe_energies = list()

    def iterate(self):
        """Performs the SA-OO-VQE iterations.

        Each iteration, a SA-VQE minimization is performed followed by an orbital optimization. This process repeats until
        max_cycles are reached or the change in energy is less than tol.
        """
        for iter in range(self.max_cycles):
            vqe_energy = self.simulate()
            self.vqe_energies.append(vqe_energy)
            self.rdms = list()
            for reference_circuit in self.reference_circuits:
                self.rdms.append(self.get_rdm(self.optimal_var_params, ref_state=reference_circuit))
            energy_new = self.energy_from_rdms()
            if self.verbose:
                print(f"The State-Averaged VQE energy for iteration {iter} is: {energy_new}")
            if iter > 0 and abs(energy_new-self.energies[-1]) < self.tol:
                self.energies.append(energy_new)
                break
            for _ in range(self.n_oo_per_iter):
                u_mat = self.generate_oo_unitary()
                self.molecule.mean_field.mo_coeff = self.molecule.mean_field.mo_coeff @ u_mat
            self.energies.append(self.energy_from_rdms())
            if self.verbose:
                print(f"The State-Averaged Orbital Optimized energy for iteration {iter} is: {self.energies[-1]}")
            self.build()

    def energy_from_rdms(self):
        "Calculate energy from rdms generated from SA_VQESolver"
        fcore, foneint, ftwoint = self.molecule.get_full_space_integrals()
        ftwoint = ftwoint.transpose(0, 3, 1, 2)
        occupied_indices = self.molecule.frozen_occupied
        active_indices = self.molecule.active_mos
        # Determine core constant
        core_constant = 0
        for i in occupied_indices:
            core_constant += 2 * foneint[i, i]
            for j in occupied_indices:
                core_constant += (2 * ftwoint[i, i, j, j] - ftwoint[i, j, i, j])

        active_energy = 0
        v_mat = np.zeros((foneint.shape[0], foneint.shape[0]))
        for t, u in product(active_indices, repeat=2):
            for i in occupied_indices:
                v_mat[u, t] += 2 * ftwoint[i, i, t, u] - ftwoint[i, t, i, u]

        n_active_mos = self.molecule.n_active_mos
        one_rdm = np.zeros((n_active_mos, n_active_mos))
        two_rdm = np.zeros((n_active_mos, n_active_mos, n_active_mos, n_active_mos))
        for i in range(self.n_ref_states):
            one_rdm += self.rdms[i][0].real*self.weights[i]
            two_rdm += self.rdms[i][1].real*self.weights[i]/2

        for ti, t in enumerate(active_indices):
            for ui, u in enumerate(active_indices):
                active_energy += one_rdm[ti, ui] * (foneint[t, u] + v_mat[t, u])
                for vi, v in enumerate(active_indices):
                    for wi, w in enumerate(active_indices):
                        active_energy += two_rdm[ti, ui, vi, wi] * ftwoint[t, u, v, w]

        return fcore + core_constant + active_energy

    def generate_oo_unitary(self):
        """Generate the orbital optimization unitary that rotates the orbitals. It uses n_oo_per_iter Newton-Raphson steps
        with the Hessian calculated analytically.

        The unitary is generated using the method outlined in
        [1] Per E. M. Siegbahn, Jan Almlof, Anders Heiberg, and Bjorn O. Roos, "The complete active space SCF (CASSCF) method
        in a Newton-Raphson formulation with application to the HNO molecule", J. Chem. Phys. 74, 2384-2396 (1981)

        Returns:
            array: The unitary matrix that when applied to the mean-field coefficients reduces the state averaged energy
        """
        _, foneint, ftwoint = self.molecule.get_full_space_integrals()
        ftwoint = ftwoint.transpose(0, 3, 1, 2)
        n_active_mos = self.molecule.n_active_mos
        n_mos = self.molecule.n_mos
        f = list(range(n_mos))
        oc = self.molecule.frozen_occupied
        ac = self.molecule.active_mos
        un = self.molecule.frozen_virtual
        n_oc = len(oc)
        n_ac = len(ac)
        n_un = len(un)

        one_rdm = np.zeros((n_active_mos, n_active_mos))
        two_rdm = np.zeros((n_active_mos, n_active_mos, n_active_mos, n_active_mos))

        for i in range(self.n_ref_states):
            one_rdm += self.rdms[i][0].real*self.weights[i]
            two_rdm += self.rdms[i][1].real*self.weights[i]/2

        # The following calculation of the analytic Hessian and gradient are derived from [1]
        f_mat = np.zeros((n_mos, n_mos))
        fi_mat = foneint.copy()
        fi_mat += 2*np.einsum("ijkk->ij", ftwoint[np.ix_(f, f, oc, oc)])-np.einsum("ikjk->ij", ftwoint[np.ix_(f, oc, f, oc)])
        fa_mat = np.einsum("tu,pqtu->pq", one_rdm, ftwoint[np.ix_(f, f, ac, ac)]) - 1/2*np.einsum("tu,ptqu->pq", one_rdm, ftwoint[np.ix_(f, ac, f, ac)])

        inds = np.ix_(oc, f)
        f_mat[inds] = 2*(fa_mat[inds]+fi_mat[inds])
        f_mat[np.ix_(ac, f)] += (np.einsum("tu,qu->tq", one_rdm, fi_mat[np.ix_(f, ac)])
                                 + 2*np.einsum("tuvx,quvx->tq", two_rdm, ftwoint[np.ix_(f, ac, ac, ac)]))

        d2ed2x = np.zeros((n_mos, n_mos, n_mos, n_mos))

        inds = np.ix_(oc, ac, oc, ac)
        indsoo = np.ix_(oc, oc)
        eye_m_one_rdm = np.eye(one_rdm.shape[0]) - one_rdm
        ftwointaoao = ftwoint[np.ix_(ac, oc, ac, oc)]
        ftwointaaoo = ftwoint[np.ix_(ac, ac, oc, oc)]
        d2ed2x[inds] += (2*(np.einsum("utvx,vxij->itju", two_rdm, ftwointaaoo)
                         + np.einsum("uxvt,vixj->itju", two_rdm + two_rdm.transpose([0, 1, 3, 2]), ftwointaoao))
                         + np.einsum("tv,viuj->itju", eye_m_one_rdm, 4*ftwointaoao-ftwointaoao.transpose([2, 1, 0, 3]) -
                                     ftwointaaoo.transpose([1, 2, 0, 3]))
                         + np.einsum("uv,vjti->itju", eye_m_one_rdm, 4*ftwointaoao-ftwointaoao.transpose([2, 1, 0, 3]) -
                                     ftwointaaoo.transpose([1, 3, 0, 2]))
                         + np.einsum("tu,ij->itju", one_rdm, fi_mat[indsoo]))
        indsaa = np.ix_(ac, ac)
        for i in oc:
            d2ed2x[np.ix_([i], ac, [i], ac)] += (2*fi_mat[indsaa]+2*fa_mat[indsaa]-f_mat[indsaa]).reshape((1, n_ac, 1, n_ac))
        for t in ac:
            d2ed2x[np.ix_(oc, [t], oc, [t])] -= 2*(fi_mat[indsoo] + fa_mat[indsoo]).reshape((n_oc, 1, n_oc, 1))

        inds = np.ix_(oc, ac, oc, un)
        eye_m_one_rdm = 2*np.eye(one_rdm.shape[0]) - one_rdm
        d2ed2x[inds] = np.einsum("tv,ajvi->itja", eye_m_one_rdm, 4*ftwoint[np.ix_(un, oc, ac, oc)] - ftwoint[np.ix_(un, ac, oc, oc)].transpose([0, 3, 1, 2])
                                 - ftwoint[np.ix_(un, oc, ac, oc)].transpose([0, 3, 2, 1]))
        indsua = np.ix_(un, ac)
        for i in oc:
            d2ed2x[np.ix_([i], ac, [i], un)] += (2*(fi_mat[indsua] + fa_mat[indsua]).transpose() - 1/2*f_mat[np.ix_(ac, un)]).reshape([1, n_ac, 1, n_un])

        inds = np.ix_(oc, ac, ac, un)
        ftwointuoaa = ftwoint[np.ix_(un, oc, ac, ac)]
        ftwointuaao = ftwoint[np.ix_(un, ac, ac, oc)]
        d2ed2x[inds] += ((-2)*(np.einsum("tuvx,aivx->itua", two_rdm, ftwointuoaa)
                         + np.einsum("tvux,axvi->itua", two_rdm+two_rdm.transpose([0, 1, 3, 2]), ftwointuaao))
                         + np.einsum("uv,avti->itua", one_rdm, 4*ftwointuaao - ftwointuoaa.transpose([0, 3, 2, 1]) -
                                     ftwointuaao.transpose([0, 2, 1, 3]))
                         - np.einsum("tu,ai->itua", one_rdm, fi_mat[np.ix_(un, oc)]))
        for t in ac:
            d2ed2x[np.ix_(oc, [t], [t], un)] += (fi_mat[np.ix_(un, oc)] + fa_mat[np.ix_(un, oc)]).transpose().reshape([n_oc, 1, 1, n_un])

        d2ed2x[np.ix_(oc, un, oc, un)] = 2*(4*ftwoint[np.ix_(un, oc, un, oc)].transpose([1, 0, 3, 2])
                                            - ftwoint[np.ix_(un, un, oc, oc)].transpose([2, 0, 3, 1])
                                            - ftwoint[np.ix_(un, oc, un, oc)].transpose([3, 0, 1, 2]))
        for i in oc:
            d2ed2x[np.ix_([i], un, [i], un)] += 2*(fi_mat[np.ix_(un, un)] + fa_mat[np.ix_(un, un)]).reshape([1, n_un, 1, n_un])
        for a in un:
            d2ed2x[np.ix_(oc, [a], oc, [a])] -= 2*(fi_mat[indsoo] + fa_mat[indsoo]).reshape([n_oc, 1, n_oc, 1])

        inds = np.ix_(oc, un, ac, un)
        d2ed2x[inds] = np.einsum("tv,aibv->iatb", one_rdm, 4*ftwoint[np.ix_(un, oc, un, ac)] -
                                 ftwoint[np.ix_(un, ac, un, oc)].transpose([0, 3, 2, 1]) -
                                 ftwoint[np.ix_(un, un, ac, oc)].transpose([0, 3, 1, 2]))
        for a in un:
            d2ed2x[np.ix_(oc, [a], ac, [a])] -= (fi_mat[np.ix_(ac, oc)] + fa_mat[np.ix_(ac, oc)] +
                                                 1/2*f_mat[np.ix_(ac, oc)]).transpose().reshape([n_oc, 1, n_ac, 1])

        inds = np.ix_(ac, un, ac, un)
        d2ed2x[inds] = (2*(np.einsum("tuvx,abvx->taub", two_rdm, ftwoint[np.ix_(un, un, ac, ac)]) +
                           np.einsum("txvu,axbv->taub", two_rdm+two_rdm.transpose([0, 1, 3, 2]), ftwoint[np.ix_(un, ac, un, ac)])) +
                        np.einsum("tu,ab->taub", one_rdm, fi_mat[np.ix_(un, un)]))
        ftwointaaaa = ftwoint[np.ix_(ac, ac, ac, ac)]
        fi_mataa = fi_mat[np.ix_(ac, ac)]
        for a in un:
            d2ed2x[np.ix_(ac, [a], ac, [a])] -= (np.einsum("tvxy,uvxy->tu", two_rdm, ftwointaaaa) +
                                                 np.einsum("uvxy,tvxy->tu", two_rdm, ftwointaaaa) +
                                                 1/2*(np.einsum("tv,uv->tu", one_rdm, fi_mataa) +
                                                      np.einsum("uv,tv->tu", one_rdm, fi_mataa))).reshape([n_ac, 1, n_ac, 1])

        ivals = oc + ac
        jvals = ac + un
        ij_list = list()
        for i in ivals:
            for j in jvals:
                if (j > i and not (i in ac and j in ac)):
                    ij_list.append([i, j])

        n_params = len(ij_list)
        hess = np.zeros((n_params, n_params))
        dedx = np.zeros(n_params)
        for p1, (i, j) in enumerate(ij_list):
            dedx[p1] = 2*(f_mat[i, j]-f_mat[j, i])
            for p2, (k, ll) in enumerate(ij_list):
                hess[p1, p2] = d2ed2x[i, j, k, ll]*2

        # Regularization to ensure all hessian eigenvalues are greater than zero
        E, _ = np.linalg.eigh(hess)
        fac = abs(E[0])*2 if E[0] < 0 else 0
        hess = hess + np.eye(n_params)*fac

        # Generate matrix elements for generating the unitary and calculate exponential of Skew-Hermitian matrix (a unitary)
        knew = -np.linalg.solve(hess, dedx)

        mat_rep = np.zeros((n_mos,  n_mos))
        for p1, (i, j) in enumerate(ij_list):
            mat_rep[i, j] = knew[p1]
            mat_rep[j, i] = -knew[p1]

        return expm(-mat_rep)



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provide useful functions, corresponding to common patterns in quantum
chemistry circuits (CNOT ladders, Pauli-word to circuit translation ...) to
facilitate the assembly of ansatz quantum circuits.
"""

from copy import deepcopy
from itertools import combinations

import numpy as np
from openfermion.ops import FermionOperator as ofFermionOperator
from openfermion.ops import InteractionOperator as ofInteractionOperator
from openfermion.ops import QubitOperator as ofQubitOperator

from tangelo.linq import Circuit, Gate
from tangelo.toolboxes.operators import FermionOperator, QubitOperator
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping, get_fermion_operator


def pauli_op_to_gate(index, op, inverse=False):
    """Return the change-of-basis gates required to map pauli words to quantum
    circuit as per Whitfield 2010 (https://arxiv.org/pdf/1001.3855.pdf).
    """
    if op == "X":
        return Gate("H", index)
    elif op == "Y":
        gate = Gate("RX", index, parameter=0.5*np.pi)
        return gate if not inverse else gate.inverse()


def exp_pauliword_to_gates(pauli_word, coef, variational=True, control=None):
    """Generate a list of Gate objects corresponding to the exponential of a pauli word.
    The process is described in Whitfield 2010 https://arxiv.org/pdf/1001.3855.pdf

    Args:
        pauli_word (tuple): Openfermion-like tuple that generates a pauli_word to exponentiate
        coef (float): The coefficient in the exponentiation
        variational (bool): When creating the Gate objects, label the (controlled-)Rz gate as variational
        control (integer): The control qubit label

    Returns:
        list: list of Gate objects that represents the exponentiation of the pauli word.
    """
    gates = []

    # Before CNOT ladder
    for index, op in pauli_word:
        if op in {"X", "Y"}:
            gates += [pauli_op_to_gate(index, op, inverse=False)]

    # CNOT ladder and rotation
    indices = sorted([index for index, op in pauli_word])
    cnot_ladder_gates = [Gate("CNOT", target=pair[1], control=pair[0]) for pair in zip(indices[:-1], indices[1:])]
    gates += cnot_ladder_gates

    angle = 2.*coef if coef >= 0. else 4*np.pi+2*coef
    if control is None:
        gates += [Gate("RZ", target=indices[-1], parameter=angle, is_variational=variational)]
    else:
        gates += [Gate("CRZ", target=indices[-1], control=control, parameter=angle)]

    gates += cnot_ladder_gates[::-1]

    # After CNOT ladder
    for index, op in pauli_word[::-1]:
        if op in {"X", "Y"}:
            gates += [pauli_op_to_gate(index, op, inverse=True)]

    return gates


def get_exponentiated_qubit_operator_circuit(qubit_op, time=1., variational=False, trotter_order=1, control=None,
                                             return_phase=False, pauli_order=None):
    """Generate the exponentiation of a qubit operator in first- or second-order Trotterized form.
    The algorithm is described in Whitfield 2010 https://arxiv.org/pdf/1001.3855.pdf

    Args:
        qubit_op  (QubitOperator):  qubit hamiltonian to exponentiate
        time (float or dict): The time to evolve the whole system or individiual times for each
            term in the operator. If a dictionary, must have keys that have a matching key in qubit_op.terms
        variational (bool) : Whether the coefficients are variational
        trotter_order (int): order of trotter approximation, only 1 or 2 are supported.
        return_phase (bool): Return the global-phase generated
        pauli_order (list): The desired pauli_word order for trotterization defined as a list of (pauli_word, coeff)
            elements which have matching dictionary elements pauli_word: coeff in QubitOperator terms.items().
            The coeff in pauli_order is used to generate the exponential.

    Returns:
        Circuit: circuit corresponding to exponentiation of qubit operator
        phase : The global phase of the time evolution if return_phase=True else not included
    """
    if pauli_order is None:
        pauli_words = list(qubit_op.terms.items())
    elif isinstance(pauli_order, list):
        pauli_words = pauli_order.copy()
    else:
        raise ValueError("ordered terms must be a list with elements (keys, values) of qubit_op.terms.items()")

    if trotter_order > 2:
        raise ValueError(f"Trotter order of >2 is not supported currently in Tangelo.")
    prefactor = 1/2 if trotter_order == 2 else 1

    if isinstance(time, (float, np.floating, np.integer, int)):
        evolve_time = {term: prefactor*time for term in qubit_op.terms.keys()}
    elif isinstance(time, dict):
        if time.keys() == qubit_op.terms.keys():
            evolve_time = {term: prefactor*etime for term, etime in time.items()}
        else:
            raise ValueError(f"The keys in the time dictionary do not match the keys in qubit_op.terms")
    else:
        raise ValueError(f"time must be a float or a dictionary")

    phase = 1.
    exp_pauli_word_gates = list()
    for i in range(trotter_order):
        if i == 1:
            pauli_words.reverse()
        for pauli_word, coef in pauli_words:
            if pauli_word:  # identity terms do not contribute to evolution outside of a phase
                if abs(np.real(coef)*evolve_time[pauli_word]) > 1.e-10:
                    exp_pauli_word_gates += exp_pauliword_to_gates(pauli_word,
                                                                   np.real(coef)*evolve_time[pauli_word],
                                                                   variational=variational,
                                                                   control=control)
            else:
                if control is None:
                    phase *= np.exp(-1j * coef * evolve_time[pauli_word])
                else:
                    exp_pauli_word_gates += [Gate("PHASE", target=control, parameter=-np.real(coef)*evolve_time[pauli_word])]

    return_value = (Circuit(exp_pauli_word_gates), phase) if return_phase else Circuit(exp_pauli_word_gates)
    return return_value


def trotterize(operator, time=1., n_trotter_steps=1, trotter_order=1, variational=False,
               mapping_options=dict(), control=None, return_phase=False):
    """Generate the circuit that represents time evolution of an operator.
    This circuit is generated as a trotterization of a qubit operator which is either the input
    or mapped from the given fermion operator.

    Args:
        operator  (QubitOperator or FermionOperator):  operator to time evolve
        time (float or dict): The time to evolve the whole system or individiual times for each
            term in the operator. If a dict, each key must match the keys in operator.terms
        variational (bool): whether the coefficients are variational
        trotter_order (int): order of trotter approximation, 1 or 2 supported
        n_trotter_steps (int): The number of different time steps taken for total time t
        mapping_options (dict): Defines the desired Fermion->Qubit mapping
                                Default values:{"up_then_down": False, "qubit_mapping": "jw", "n_spinorbitals": None,
                                                "n_electrons": None}
        control (int): The label for the control Qubit of the time-evolution
        return_phase (bool): If return_phase is True, the global phase of the time-evolution will be returned

    Returns:
        Circuit: circuit corresponding to time evolution of the operator
        float: the global phase not included in the circuit if return_phase=True else not included
    """

    if isinstance(operator, ofFermionOperator):
        options = {"up_then_down": False, "qubit_mapping": "jw", "n_spinorbitals": None, "n_electrons": None}
        # Overwrite default values with user-provided ones, if they correspond to a valid keyword
        for k, v in mapping_options.items():
            if k in options:
                options[k] = v
            else:
                raise KeyError(f"Keyword :: {k}, not a valid fermion to qubit mapping option")
        if isinstance(time, (float, np.floating, int, np.integer)):
            evolve_time = {term: time for term in operator.terms.keys()}
        elif isinstance(time, dict):
            if time.keys() == operator.terms.keys():
                evolve_time = deepcopy(time)
            else:
                raise ValueError(f"keys of time do not match keys of operator.terms")
        else:
            raise ValueError("time must be a float or dictionary of floats")
        new_operator = FermionOperator()
        for term in operator.terms:
            new_operator += FermionOperator(term, operator.terms[term]*evolve_time[term]/n_trotter_steps)
        qubit_op = fermion_to_qubit_mapping(fermion_operator=new_operator,
                                            mapping=options["qubit_mapping"],
                                            n_spinorbitals=options["n_spinorbitals"],
                                            n_electrons=options["n_electrons"],
                                            up_then_down=options["up_then_down"])
        circuit, phase = get_exponentiated_qubit_operator_circuit(qubit_op,
                                                                  time=1.,  # time is already included
                                                                  trotter_order=trotter_order,
                                                                  variational=variational,
                                                                  control=control,
                                                                  return_phase=True)

    elif isinstance(operator, (ofQubitOperator)):
        qubit_op = deepcopy(operator)
        if isinstance(time, float):
            evolve_time = time / n_trotter_steps
        elif isinstance(time, dict):
            if time.keys() == operator.terms.keys():
                evolve_time = {term: etime / n_trotter_steps for term, etime in time.items()}
            else:
                raise ValueError(f"time dictionary and operator.terms dictionary have different keys.")
        else:
            raise ValueError(f"time must be a float or a dictionary of floats")
        circuit, phase = get_exponentiated_qubit_operator_circuit(qubit_op,
                                                                  time=evolve_time,
                                                                  trotter_order=trotter_order,
                                                                  variational=variational,
                                                                  control=control,
                                                                  return_phase=True)
    else:
        raise ValueError("Only FermionOperator or QubitOperator allowed")

    return_value = (circuit*n_trotter_steps, phase**n_trotter_steps) if return_phase else circuit*n_trotter_steps
    return return_value


def append_qft_rotations_gates(gate_list, qubit_list, prefac=1):
    """Appends the list of gates required for a quantum fourier transform to a gate list.

    Args:
        gate_list (list): List of Gate elements
        qubit_list (list): List of integers for which the qft operations are performed

    Returns:
        list: List of Gate objects for rotation portion of qft circuit appended to gate_list"""
    n = len(qubit_list)
    if n == 0:
        return gate_list
    n -= 1
    gate_list += [Gate("H", target=qubit_list[n])]
    for i, qubit in enumerate(qubit_list[:n]):
        gate_list += [Gate("CPHASE", control=qubit, target=qubit_list[n], parameter=prefac*np.pi/2**(n-i))]

    append_qft_rotations_gates(gate_list, qubit_list[:n], prefac=prefac)


def swap_registers(gate_list, qubit_list):
    """Function to swap register order.
    Args:
        gate_list (list): List of Gate
        qubit_list (list): List of integers for the locations of the qubits

    Result:
        list: The Gate operations that swap the register order"""
    n = len(qubit_list)
    for qubit_index in range(n//2):
        gate_list += [Gate("SWAP", target=[qubit_list[qubit_index], qubit_list[n - qubit_index - 1]])]
    return gate_list


def get_qft_circuit(qubits, n_qubits=None, inverse=False, swap=True):
    """Returns the QFT or iQFT circuit given a list of qubits to act on.

    Args:
        qubits (int or list): The list of qubits to apply the QFT circuit to. If an integer,
            the operation is applied to the [0,...,qubits-1] qubits
        n_qubits: Argument to initialize a Circuit with the desired number of qubits.
        inverse (bool): If True, the inverse QFT is applied. If False, QFT is applied
        swap (bool): Whether to apply swap to the registers.

        Returns:
            Circuit: The circuit that applies QFT or iQFT to qubits
        """

    if isinstance(qubits, int):
        qubit_list = list(range(qubits))
    elif isinstance(qubits, list):
        qubit_list = qubits
    else:
        raise KeyError("qubits must be an int or list of ints")

    swap_gates = list()
    if swap:
        swap_registers(swap_gates, qubit_list)

    qft_gates = list()
    if inverse:
        append_qft_rotations_gates(qft_gates, qubit_list, prefac=-1)
        qft_gates = [gate for gate in reversed(qft_gates)]
        qft_gates = swap_gates + qft_gates
    else:
        append_qft_rotations_gates(qft_gates, qubit_list)
        qft_gates += swap_gates

    return Circuit(qft_gates, n_qubits=n_qubits)


def controlled_pauliwords(qubit_op, control, n_qubits=None):
    """Takes a qubit operator and returns controlled-pauliword circuits for each term as a list.

    Args:
        qubit_op (QubitOperator): The qubit operator with pauliwords to generate circuits for
        control (int): The index of the control qubit
        n_qubits (int): When generating each Circuit, create with n_qubits size

    Returns:
        list: List of controlled-pauliword Circuit for each pauliword in the qubit_op
    """
    pauli_words = qubit_op.terms.items()

    pauliword_circuits = list()
    for (pauli_word, _) in pauli_words:
        gates = [Gate(name="C"+op, target=index, control=control) for index, op in pauli_word]
        pauliword_circuits.append(Circuit(gates, n_qubits=n_qubits))
    return pauliword_circuits


def controlled_swap_to_XX_gates(c, n1, n2):
    """Equivalent decomposition of controlled swap into 1-qubit gates and XX 2-qubit gate.

    This is useful for IonQ experiments as the native two-qubit gate is the XX Ising coupling.

    Args:
        c (int): control qubit
        n1 (int): first target qubit
        n2 (int): second target qubit

    Returns:
        list: List of Gate that applies controlled swap operation
    """
    gates = [Gate("RY", target=c, parameter=7*np.pi/2.),
             Gate("RZ", target=n1, parameter=7*np.pi/2.),
             Gate("XX", target=[n1, n2], parameter=5*np.pi/2.),
             Gate("RZ", target=n1, parameter=7*np.pi/4.),
             Gate("RZ", target=n2, parameter=3*np.pi/4.),
             Gate("RY", target=n1, parameter=np.pi/2.),
             Gate("XX", target=[c, n2], parameter=7*np.pi/2.),
             Gate("RY", target=n2, parameter=11*np.pi/4),
             Gate("XX", target=[n1, n2], parameter=7*np.pi/2.),
             Gate("XX", target=[c, n1], parameter=np.pi/4.),
             Gate("RZ", target=n2, parameter=np.pi/4),
             Gate("XX", target=[c, n2], parameter=5*np.pi/2),
             Gate("RY", target=c, parameter=5*np.pi/2),
             Gate("RZ", target=n1, parameter=5*np.pi/2),
             Gate("RY", target=n2, parameter=7*np.pi/4),
             Gate("XX", target=[n1, n2], parameter=7*np.pi/2),
             Gate("RY", target=n1, parameter=np.pi/2),
             Gate("RZ", target=c, parameter=11*np.pi/4)]
    return gates


def derangement_circuit(qubit_list, control=None, n_qubits=None, decomp=None):
    """Returns the (controlled-)derangement circuit for multiple copies of a state

    Args:
        qubit_list (list of list(int)): Each item in the list is a list of qubit registers for each copy. The length of
            each list of qubit registers must be the same.
            For example [[1, 2], [3, 4]] applies controlled-swaps between equivalent states located on qubits [1, 2] and [3, 4]
        control (int): The control register to be measured.
        n_qubits (int): The number of qubits in the circuit.
        decomp (str): Use the decomposed controlled-swap into 1-qubit gates and a certain 2-qubit gate listed below.
            "XX": 2-qubit gate is XX

    Returns:
        Circuit: The derangement circuit
    """
    if decomp is not None and decomp not in ["XX"]:
        raise ValueError(f"{decomp} is not a valid controlled swap decomposition")

    num_copies = len(qubit_list)
    if num_copies == 1:
        return Circuit(n_qubits=n_qubits)
    else:
        rho_range = len(qubit_list[0])
        for i in range(1, num_copies):
            if len(qubit_list[i]) != rho_range:
                raise ValueError("All copies must have the same number of qubits")
    gate_list = list()
    if control is None:
        for copy1, copy2 in combinations(range(num_copies), 2):
            for rhoi in range(rho_range):
                gate_list += [Gate("SWAP", target=[qubit_list[copy1][rhoi], qubit_list[copy2][rhoi]])]
    else:
        for copy1, copy2 in combinations(range(num_copies), 2):
            for rhoi in range(rho_range):
                if decomp == "XX":
                    gate_list += controlled_swap_to_XX_gates(control,
                                                             qubit_list[copy1][rhoi],
                                                             qubit_list[copy2][rhoi])
                else:
                    gate_list += [Gate("CSWAP",
                                       target=[qubit_list[copy1][rhoi], qubit_list[copy2][rhoi]],
                                       control=control)]

    return Circuit(gate_list, n_qubits=n_qubits)


def givens_gate(target, theta, is_variational=False):
    """Generates the list of gates corresponding to a givens rotation exp(-theta*(XX+YY))

    Explicitly the two-qubit matrix is
    [[0,      0,           0,       0],
     [0,  cos(theta), -sin(theta),  0],
     [0,  sin(theta),  cos(theta),  0],
     [0,      0,            0,      0]]

    Args:
        target (list): list of two integers that indicate which qubits are involved in the givens rotation
        theta (float): the rotation angle
        is_variational (bool): Whether the rotation angle is a variational parameter.

    Returns:
        list of Gate: The list of gates corresponding to the givens rotation"""
    if len(target) != 2:
        raise ValueError("target must be a list or array of two integers")
    return [Gate("CNOT", target=target[0], control=target[1]),
            Gate("CRY", target=target[1], control=target[0], parameter=-theta, is_variational=is_variational),
            Gate("CNOT", target=target[0], control=target[1])]



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions helping with quantum circuit format conversion between abstract
format and circ format.

In order to produce an equivalent circuit for the target backend, it is
necessary to account for:
- how the gate names differ between the source backend to the target backend.
- how the order and conventions for some of the inputs to the gate operations
    may also differ.
"""
from math import pi


def get_cirq_gates():
    """Map gate name of the abstract format to the equivalent methods of the
    cirq class API and supported gates: https://quantumai.google/cirq/gates.
    """
    import cirq

    GATE_CIRQ = dict()
    GATE_CIRQ["H"] = cirq.H
    GATE_CIRQ["X"] = cirq.X
    GATE_CIRQ["Y"] = cirq.Y
    GATE_CIRQ["Z"] = cirq.Z
    GATE_CIRQ["CX"] = cirq.X
    GATE_CIRQ["CY"] = cirq.Y
    GATE_CIRQ["CZ"] = cirq.Z
    GATE_CIRQ["S"] = cirq.S
    GATE_CIRQ["T"] = cirq.T
    GATE_CIRQ["CH"] = cirq.H
    GATE_CIRQ["RX"] = cirq.rx
    GATE_CIRQ["RY"] = cirq.ry
    GATE_CIRQ["RZ"] = cirq.rz
    GATE_CIRQ["CNOT"] = cirq.CNOT
    GATE_CIRQ["CRZ"] = cirq.rz
    GATE_CIRQ["CRX"] = cirq.rx
    GATE_CIRQ["CRY"] = cirq.ry
    GATE_CIRQ["PHASE"] = cirq.ZPowGate
    GATE_CIRQ["CPHASE"] = cirq.ZPowGate
    GATE_CIRQ["XX"] = cirq.XXPowGate
    GATE_CIRQ["SWAP"] = cirq.SWAP
    GATE_CIRQ["CSWAP"] = cirq.SWAP
    GATE_CIRQ["MEASURE"] = cirq.measure
    return GATE_CIRQ


def translate_cirq(source_circuit, noise_model=None):
    """Take in an abstract circuit, return an equivalent cirq QuantumCircuit
    instance.

    Args:
        source_circuit: quantum circuit in the abstract format.

    Returns:
        cirq.Circuit: a corresponding cirq Circuit. Right now, the structure is
            of LineQubit. It is possible in the future that we may support
            NamedQubit or GridQubit.
    """
    import cirq

    GATE_CIRQ = get_cirq_gates()
    target_circuit = cirq.Circuit()
    # cirq by definition uses labels for qubits, this is one way to automatically generate
    # labels. Could also use GridQubit for square lattice or NamedQubit to name qubits
    qubit_list = cirq.LineQubit.range(source_circuit.width)
    # Add next line to make sure all qubits are initialized
    # cirq will otherwise only initialize qubits that have gates
    target_circuit.append(cirq.I.on_each(qubit_list))

    # Maps the gate information properly. Different for each backend (order, values)
    for gate in source_circuit._gates:
        if (gate.control is not None) and gate.name != 'CNOT':
            num_controls = len(gate.control)
            control_list = [qubit_list[c] for c in gate.control]
        if gate.name in {"H", "X", "Y", "Z", "S", "T"}:
            target_circuit.append(GATE_CIRQ[gate.name](qubit_list[gate.target[0]]))
        elif gate.name in {"CH", "CX", "CY", "CZ"}:
            next_gate = GATE_CIRQ[gate.name].controlled(num_controls)
            target_circuit.append(next_gate(*control_list, qubit_list[gate.target[0]]))
        elif gate.name in {"RX", "RY", "RZ"}:
            next_gate = GATE_CIRQ[gate.name](gate.parameter)
            target_circuit.append(next_gate(qubit_list[gate.target[0]]))
        elif gate.name in {"CNOT"}:
            target_circuit.append(GATE_CIRQ[gate.name](qubit_list[gate.control[0]], qubit_list[gate.target[0]]))
        elif gate.name in {"MEASURE"}:
            target_circuit.append(GATE_CIRQ[gate.name](qubit_list[gate.target[0]]))
        elif gate.name in {"CRZ", "CRX", "CRY"}:
            next_gate = GATE_CIRQ[gate.name](gate.parameter).controlled(num_controls)
            target_circuit.append(next_gate(*control_list, qubit_list[gate.target[0]]))
        elif gate.name in {"XX"}:
            next_gate = GATE_CIRQ[gate.name](exponent=gate.parameter/pi, global_shift=-0.5)
            target_circuit.append(next_gate(qubit_list[gate.target[0]], qubit_list[gate.target[1]]))
        elif gate.name in {"PHASE"}:
            next_gate = GATE_CIRQ[gate.name](exponent=gate.parameter/pi)
            target_circuit.append(next_gate(qubit_list[gate.target[0]]))
        elif gate.name in {"CPHASE"}:
            next_gate = GATE_CIRQ[gate.name](exponent=gate.parameter/pi).controlled(num_controls)
            target_circuit.append(next_gate(*control_list, qubit_list[gate.target[0]]))
        elif gate.name in {"SWAP"}:
            target_circuit.append(GATE_CIRQ[gate.name](qubit_list[gate.target[0]], qubit_list[gate.target[1]]))
        elif gate.name in {"CSWAP"}:
            next_gate = GATE_CIRQ[gate.name].controlled(num_controls)
            target_circuit.append(next_gate(*control_list, qubit_list[gate.target[0]], qubit_list[gate.target[1]]))
        else:
            raise ValueError(f"Gate '{gate.name}' not supported on backend cirq")

        # Add noisy gates
        if noise_model and (gate.name in noise_model.noisy_gates):
            for nt, np in noise_model._quantum_errors[gate.name]:
                if nt == 'pauli':
                    # Define pauli gate in cirq language
                    depo = cirq.asymmetric_depolarize(np[0], np[1], np[2])

                    target_circuit += [depo(qubit_list[t]) for t in gate.target]
                    if gate.control is not None:
                        target_circuit += [depo(qubit_list[c]) for c in gate.control]
                elif nt == 'depol':
                    depo_list = [qubit_list[t] for t in gate.target]
                    if gate.control is not None:
                        depo_list += [qubit_list[c] for c in gate.control]
                    depo_size = len(depo_list)
                    # define depo_size-qubit depolarization gate
                    depo = cirq.depolarize(np*(4**depo_size-1)/4**depo_size, depo_size)  # param, num_qubits
                    target_circuit.append(depo(*depo_list))  # gates targeted

    return target_circuit



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Module to remove two qubits from the problem space using conservation of
electron number and conservation of electron spin. As described in
arXiv:1701.08213 and Phys. Rev. X 6, 031007.
"""


import numpy as np
import copy

from openfermion import FermionOperator as ofFermionOperator
from openfermion.transforms import bravyi_kitaev_tree, reorder
from openfermion.utils import count_qubits
from openfermion.utils import up_then_down as up_then_down_order


def symmetry_conserving_bravyi_kitaev(fermion_operator, n_spinorbitals,
                                      n_electrons, up_then_down=False, spin=0):
    """Returns the QubitOperator for the FermionOperator supplied, with two
    qubits removed using conservation of (parity) of electron spin and number,
    as described in arXiv:1701.08213.  This function has been modified from its
    analogous implementation in openfermion in order to circumvent failures when
    passing a fermion_operator which does not explicitly reference the highest
    index qubit in the register.

    Args:
        fermion_operator (FermionOperator): fermionic operator to transform to
            QubitOperator.
        n_spinorbitals (int): The number of active spin-orbitals being
            considered for the system.
        n_electrons (int): The number of active fermions being considered for
            the system (note, this is less than the number of electrons in a
            molecule if some orbitals have been frozen).
        up_then_down (bool): specify if the spin-orbital basis is already
            ordered putting all spin up before all spin down states.

    Returns:
        QubitOperator: The qubit operator corresponding to the supplied
            fermionic operator, with two qubits removed using spin symmetries.

    WARNING:
        Reorders orbitals from the default even-odd ordering to all spin-up
        orbitals, then all spin-down orbitals.

    Raises:
        ValueError if fermion_hamiltonian isn"t of the type FermionOperator, or
        active_orbitals isn"t an integer, or active_fermions isn"t an integer.

    Notes: This function reorders the spin orbitals as all spin-up, then all
        spin-down. It uses the OpenFermion bravyi_kitaev_tree mapping, rather
        than the bravyi-kitaev mapping. Caution advised when using with a
        Fermi-Hubbard Hamiltonian; this technique correctly reduces the
        Hamiltonian only for the lowest energy even and odd fermion number
        states, not states with an arbitrary number of fermions.
    """
    if not isinstance(fermion_operator, ofFermionOperator):
        raise ValueError("Supplied operator should be an instance "
                         "of openfermion FermionOperator class.")
    if type(n_spinorbitals) is not int:
        raise ValueError("Number of spin-orbitals should be an integer.")
    if type(n_electrons) is not int:
        raise ValueError("Number of electrons should be an integer.")
    if n_spinorbitals < count_qubits(fermion_operator):
        raise ValueError("Number of spin-orbitals is too small for FermionOperator input.")
    # Check that the input operator is suitable for application of scBK
    check_operator(fermion_operator, num_orbitals=(n_spinorbitals//2), up_then_down=up_then_down)

    # If necessary, arrange spins up then down, then BK map to qubit Hamiltonian.
    if not up_then_down:
        fermion_operator = reorder(fermion_operator, up_then_down_order, num_modes=n_spinorbitals)
    qubit_operator = bravyi_kitaev_tree(fermion_operator, n_qubits=n_spinorbitals)
    qubit_operator.compress()

    n_alpha = n_electrons//2 + spin//2 + (n_electrons % 2)

    # Allocates the parity factors for the orbitals as in arXiv:1704.08213.
    parity_final_orb = (-1)**n_electrons
    parity_middle_orb = (-1)**n_alpha

    # Removes the final qubit, then the middle qubit.
    qubit_operator = edit_operator_for_spin(qubit_operator,
                                            n_spinorbitals,
                                            parity_final_orb)
    qubit_operator = edit_operator_for_spin(qubit_operator,
                                            n_spinorbitals/2,
                                            parity_middle_orb)

    # We remove the N/2-th and N-th qubit from the register.
    to_prune = (n_spinorbitals//2 - 1, n_spinorbitals - 1)
    qubit_operator = prune_unused_indices(qubit_operator, prune_indices=to_prune, n_qubits=n_spinorbitals)

    return qubit_operator


def edit_operator_for_spin(qubit_operator, spin_orbital, orbital_parity):
    """Removes the Z terms acting on the orbital from the operator. For qubits
    to be tapered out, the action of Z-operators in operator terms are reduced
    to the associated eigenvalues. This simply corresponds to multiplying term
    coefficients by the related eigenvalue +/-1.

    Args:
        qubit_operator (QubitOperator): input operator.
        spin_orbital (int): index of qubit encoding (spin/occupation) parity.
        orbital_parity (int): plus/minus one, parity of eigenvalue.

    Returns:
        QubitOperator: updated operator, with relevant coefficients multiplied
            by +/-1.
    """
    new_qubit_dict = {}
    for term, coefficient in qubit_operator.terms.items():
        # If Z acts on the specified orbital, precompute its effect and
        # remove it from the Hamiltonian.
        if (spin_orbital - 1, "Z") in term:
            new_coefficient = coefficient*orbital_parity
            new_term = tuple(i for i in term if i != (spin_orbital - 1, "Z"))
            # Make sure to combine terms comprised of the same operators.
            if new_qubit_dict.get(new_term) is None:
                new_qubit_dict[new_term] = new_coefficient
            else:
                old_coefficient = new_qubit_dict.get(new_term)
                new_qubit_dict[new_term] = new_coefficient + old_coefficient
        else:
            # Make sure to combine terms comprised of the same operators.
            if new_qubit_dict.get(term) is None:
                new_qubit_dict[term] = coefficient
            else:
                old_coefficient = new_qubit_dict.get(term)
                new_qubit_dict[term] = coefficient + old_coefficient

    qubit_operator.terms = new_qubit_dict
    qubit_operator.compress()

    return qubit_operator


def prune_unused_indices(qubit_operator, prune_indices, n_qubits):
    """Rewritten from openfermion implementation. This uses the number of qubits,
    rather than the operator itself to specify the number of qubits relevant to
    the problem. This is especially important for, e.g. terms in the ansatz
    which may not individually pertain to all qubits in the problem.

    Remove indices that do not appear in any terms.

    Indices will be renumbered such that if an index i does not appear in any
    terms, then the next largest index that appears in at least one term will be
    renumbered to i.

    Args:
        qubit_operator (QubitOperator): input operator.
        prune_indices (tuple of int): indices to be removed from qubit register.
        n_qubits (int): number of qubits in register.

    Returns:
        QubitOperator: output operator, with designated qubit indices excised.
    """

    indices = np.linspace(0, n_qubits - 1, n_qubits, dtype=int)
    indices = np.delete(indices, prune_indices)

    # Construct a dict that maps the old indices to new ones
    index_map = {}
    for index in enumerate(indices):
        index_map[index[1]] = index[0]

    new_operator = copy.deepcopy(qubit_operator)
    new_operator.terms.clear()

    # Replace the indices in the terms with the new indices
    for term in qubit_operator.terms:
        new_term = [(index_map[op[0]], op[1]) for op in term]
        new_operator.terms[tuple(new_term)] = qubit_operator.terms[term]

    return new_operator


def check_operator(fermion_operator, num_orbitals=None, up_then_down=False):
    """Check if the input fermion operator is suitable for application of
    symmetry-consering BK qubit reduction. Excitation must: preserve parity of
    fermion occupation, and parity of spin expectation value. This assumes
    alternating spin-up/spin-down ordering of input operator.

    Args:
        fermion_operator (FermionOperator): input fermionic operator.
        num_orbitals (int): specify number of orbitals (number of modes / 2),
            required for up then down ordering.
        up_then_down (bool): True if all spin up before all spin down, otherwise
            alternates.
    """
    if up_then_down and (num_orbitals is None):
        raise ValueError("Up then down spin ordering requires number of modes specified.")
    for term in fermion_operator.terms:
        number_change = 0
        spin_change = 0
        for index, action in term:
            number_change += 2*action - 1
            if up_then_down:
                spin_change += (2*action - 1)*(-2*(index // num_orbitals) + 1)*0.5

            else:
                spin_change += (2*action - 1)*(-2*(index % 2) + 1)*0.5
        if number_change % 2 != 0:
            raise ValueError("Invalid operator: input fermion operator does not conserve occupation parity.")
        if spin_change % 2 != 0:
            raise ValueError("Invalid operator: input fermion operator does not conserve spin parity.")



"""
    A test class to check that features related to the IonQ API are behaving as expected.
    Tests requiring actual interactions with the services have been skipped.
"""

import unittest
import os
import pprint

from tangelo.linq import Gate, Circuit
from tangelo.linq.qpu_connection import IonQConnection

circ1 = Circuit([Gate("H", 0), Gate("X", 1)])
res_simulator_circ1 = {'01': 0.5, '11': 0.5}


def assert_freq_dict_almost_equal(d1, d2, atol):
    """ Utility function to check whether two frequency dictionaries are almost equal, for arbitrary tolerance """
    if d1.keys() != d2.keys():
        raise AssertionError("Dictionary keys differ. Frequency dictionaries are not almost equal.\n"
                             f"d1 keys: {d1.keys()} \nd2 keys: {d2.keys()}")
    else:
        for k in d1.keys():
            if abs(d1[k] - d2[k]) > atol:
                raise AssertionError(f"Frequency {k}, difference above tolerance {atol}: {d1[k]} != {d2[k]}")


@unittest.skip("We do not want to store login information for automated testing")
class TestIonQConnection(unittest.TestCase):

    def test_init(self):
        """ If user has set environment variables IONQ_APIKEY to the correct value, this should succeed.
        Implicitly makes a call to IonQConnection.job_get_history in order to validate the apiKey. """
        IonQConnection()

    def test_init_fail(self):
        """ If user has not set environment variables IONQ_APIKEY to the correct value, this should
        return a RuntimeError. """
        tmp = os.getenv("IONQ_APIKEY", '')
        os.environ['IONQ_APIKEY'] = 'invalid_apikey'
        self.assertRaises(RuntimeError, IonQConnection)
        os.environ['IONQ_APIKEY'] = ''
        self.assertRaises(RuntimeError, IonQConnection)
        os.environ['IONQ_APIKEY'] = tmp

    def test_submit_job_simulator(self):
        """ Submit a valid job to a API validation backend (simulator) and retrieve results """

        ionq_api = IonQConnection()

        job_id = ionq_api.job_submit('simulator', circ1, 100, 'test_simulator_submit')
        job_results = ionq_api.job_results(job_id)
        pprint.pprint(job_results)

        assert_freq_dict_almost_equal(job_results, res_simulator_circ1, 1e-7)

    def test_delete_job(self):
        """ Submit a job and then cancel/delete it, regardless of its status. Check job history before and after. """

        ionq_api = IonQConnection()

        job_id = ionq_api.job_submit('simulator', circ1, 1000, 'test_simulator_cancel')
        job_history_df_before = ionq_api.job_get_history()
        assert(job_id in job_history_df_before.id.values)
        print(job_history_df_before)
        ionq_api.job_cancel(job_id)
        job_history_df_after = ionq_api.job_get_history()
        assert(job_id not in job_history_df_after.id.values)

    def test_get_backend_info(self):
        """ Retrieve backend info """
        ionq_api = IonQConnection()

        res = ionq_api.get_backend_info()
        pprint.pprint(res)

    def test_get_characterization(self):
        """ Get device characterization through name or charac url """
        ionq_api = IonQConnection()
        backend = 'qpu.s11'  # Pick something that has a charac url for this test to be useful

        res = ionq_api.get_backend_info()
        pprint.pprint(res)

        # Retrieve charac info from backend name
        d1 = ionq_api.get_characterization(backend_name=backend)

        # Retrieve charac info from charac url
        charac_url = res[res['backend'] == backend]['characterization_url'].iat[0]
        d2 = ionq_api.get_characterization(charac_url=charac_url)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines the UCCSD ansatz class. It provides a chemically inspired
ansatz and is an implementation of the classical unitary CCSD ansatz. Single and
double excitation determinants, in accordance with the system number of electron
and spin, are considered. For more information about this ansatz, see references
below.

Refs:
    - P.Kl. Barkoutsos, J.F. Gonthier, I. Sokolov, N. Moll, G. Salis, A. Fuhrer,
        M. Ganzhorn, D.J. Egger, M. Troyer, A. Mezzacapo, S. Filipp, and
        I. Tavernelli. Phys. Rev. A 98, 022322 (2018).
    - I.O. Sokolov, P.Kl. Barkoutsos, P.J. Ollitrault, D. Greenberg, J. Rice,
        M. Pistoia, and I. Tavernelli. J. Chem. Phys. 152, 124107 (2020).
    - Y. Shen, X. Zhang, S. Zhang, J.N. Zhang, M.H. Yung, and K. Kim.
        Physical Review A 95, 020501 (2017).
"""

import itertools
import numpy as np
from pyscf import mp
from openfermion.circuits import uccsd_singlet_generator

from tangelo.linq import Circuit

from .ansatz import Ansatz
from .ansatz_utils import exp_pauliword_to_gates
from ._unitary_cc_openshell import uccsd_openshell_paramsize, uccsd_openshell_generator
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping
from tangelo.toolboxes.qubit_mappings.statevector_mapping import get_reference_circuit


class UCCSD(Ansatz):
    """This class implements the UCCSD ansatz. Currently, closed-shell and
    restricted open-shell UCCSD are supported. This implies that the mean-field
    is computed with the RHF or ROHF reference integrals.

    Args:
        molecule (SecondQuantizedMolecule) : The molecular system.
        mapping (str) : one of the supported qubit mapping identifiers. Default,
            "jw".
        up_then_down (bool): change basis ordering putting all spin up orbitals
            first, followed by all spin down. Default, False (i.e. has
            alternating spin up/down ordering).
    """

    def __init__(self, molecule, mapping="JW", up_then_down=False, spin=None):

        self.molecule = molecule
        self.n_spinorbitals = molecule.n_active_sos
        self.n_electrons = molecule.n_active_electrons
        self.spin = molecule.spin if spin is None else spin
        self.mapping = mapping
        self.up_then_down = up_then_down

        # Later: refactor to handle various flavors of UCCSD
        if self.n_spinorbitals % 2 != 0:
            raise ValueError("The total number of spin-orbitals should be even.")

        # choose open-shell uccsd if spin not zero, else choose singlet ccsd
        if self.spin != 0:
            self.n_alpha = self.n_electrons//2 + self.spin//2 + 1 * (self.n_electrons % 2)
            self.n_beta = self.n_electrons//2 - self.spin//2
            self.n_singles, self.n_doubles, _, _, _, _, _ = uccsd_openshell_paramsize(self.n_spinorbitals, self.n_alpha, self.n_beta)
        else:
            self.n_spatial_orbitals = self.n_spinorbitals // 2
            self.n_occupied = int(np.ceil(self.n_electrons / 2))
            self.n_virtual = self.n_spatial_orbitals - self.n_occupied
            self.n_singles = self.n_occupied * self.n_virtual
            self.n_doubles = self.n_singles * (self.n_singles + 1) // 2

        # set total number of parameters
        self.n_var_params = self.n_singles + self.n_doubles

        # Supported reference state initialization
        # TODO: support for others
        self.supported_reference_state = {"HF", "zero"}
        # Supported var param initialization
        self.supported_initial_var_params = {"ones", "random", "mp2"} if self.spin == 0 else {"ones", "random"}

        # Default initial parameters for initialization
        # TODO: support for openshell MP2 initialization
        self.var_params_default = "mp2" if self.spin == 0 else "ones"
        self.default_reference_state = "HF"

        self.var_params = None
        self.circuit = None

    def set_var_params(self, var_params=None):
        """Set values for variational parameters, such as zeros, random numbers,
        MP2 (...), providing some keywords for users, and also supporting direct
        user input (list or numpy array). Return the parameters so that
        workflows such as VQE can retrieve these values.
        """

        if var_params is None:
            var_params = self.var_params_default

        if isinstance(var_params, str):
            var_params = var_params.lower()
            if (var_params not in self.supported_initial_var_params):
                raise ValueError(f"Supported keywords for initializing variational parameters: {self.supported_initial_var_params}")
            if var_params == "ones":
                initial_var_params = np.ones((self.n_var_params,), dtype=float)
            elif var_params == "random":
                initial_var_params = 2.e-1 * (np.random.random((self.n_var_params,)) - 0.5)
            elif var_params == "mp2":
                initial_var_params = self._compute_mp2_params()
        else:
            initial_var_params = np.array(var_params)
            if initial_var_params.size != self.n_var_params:
                raise ValueError(f"Expected {self.n_var_params} variational parameters but "
                                 f"received {initial_var_params.size}.")
        self.var_params = initial_var_params
        return initial_var_params

    def prepare_reference_state(self):
        """Returns circuit preparing the reference state of the ansatz (e.g
        prepare reference wavefunction with HF, multi-reference state, etc).
        These preparations must be consistent with the transform used to obtain
        the qubit operator.
        """

        if self.default_reference_state not in self.supported_reference_state:
            raise ValueError(f"Only supported reference state methods are:{self.supported_reference_state}")

        if self.default_reference_state == "HF":
            return get_reference_circuit(n_spinorbitals=self.n_spinorbitals,
                                         n_electrons=self.n_electrons,
                                         mapping=self.mapping,
                                         up_then_down=self.up_then_down,
                                         spin=self.spin)
        elif self.default_reference_state == "zero":
            return Circuit()

    def build_circuit(self, var_params=None):
        """Build and return the quantum circuit implementing the state
        preparation ansatz (with currently specified initial_state and
        var_params).
        """

        if var_params is not None:
            self.set_var_params(var_params)
        elif self.var_params is None:
            self.set_var_params()

        # Build qubit operator required to build UCCSD
        qubit_op = self._get_singlet_qubit_operator() if self.spin == 0 else self._get_openshell_qubit_operator()

        # Prepend reference state circuit
        reference_state_circuit = self.prepare_reference_state()

        # Obtain quantum circuit through trivial trotterization of the qubit operator
        # Keep track of the order in which pauli words have been visited for fast subsequent parameter updates
        pauli_words = sorted(qubit_op.terms.items(), key=lambda x: len(x[0]))
        pauli_words_gates = []
        self.pauli_to_angles_mapping = dict()
        for i, (pauli_word, coef) in enumerate(pauli_words):
            pauli_words_gates += exp_pauliword_to_gates(pauli_word, coef)
            self.pauli_to_angles_mapping[pauli_word] = i

        uccsd_circuit = Circuit(pauli_words_gates)
        # skip over the reference state circuit if it is empty
        if reference_state_circuit.size != 0:
            self.circuit = reference_state_circuit + uccsd_circuit
        else:
            self.circuit = uccsd_circuit

    def update_var_params(self, var_params):
        """Shortcut: set value of variational parameters in the already-built
        ansatz circuit member. Preferable to rebuilt your circuit from scratch,
        which can be an involved process.
        """

        self.set_var_params(var_params)

        # Build qubit operator required to build UCCSD
        qubit_op = self._get_singlet_qubit_operator() if self.spin == 0 else self._get_openshell_qubit_operator()

        # If qubit operator terms have changed, rebuild circuit. Else, simply update variational gates directly
        if set(self.pauli_to_angles_mapping.keys()) != set(qubit_op.terms.keys()):
            self.build_circuit(var_params)
        else:
            for pauli_word, coef in qubit_op.terms.items():
                gate_index = self.pauli_to_angles_mapping[pauli_word]
                self.circuit._variational_gates[gate_index].parameter = 2.*coef if coef >= 0. else 4*np.pi+2*coef

    def _get_singlet_qubit_operator(self):
        """Construct UCCSD FermionOperator for current variational parameters,
        and translate to QubitOperator via relevant qubit mapping.

        Returns:
            QubitOperator: qubit-encoded elements of the UCCSD ansatz.
        """
        fermion_op = uccsd_singlet_generator(self.var_params, self.n_spinorbitals, self.n_electrons)
        qubit_op = fermion_to_qubit_mapping(fermion_operator=fermion_op,
                                            mapping=self.mapping,
                                            n_spinorbitals=self.n_spinorbitals,
                                            n_electrons=self.n_electrons,
                                            up_then_down=self.up_then_down)

        # Cast all coefs to floats (rotations angles are real)
        for key in qubit_op.terms:
            qubit_op.terms[key] = float(qubit_op.terms[key].imag)
        qubit_op.compress()
        return qubit_op

    def _get_openshell_qubit_operator(self):
        """Construct open-shell UCCSD FermionOperator for current variational
        parameters, and translate to QubitOperator via relevant qubit mapping.

        Returns:
            QubitOperator: qubit-encoded elements of the UCCSD ansatz.
        """
        fermion_op = uccsd_openshell_generator(self.var_params,
                                               self.n_spinorbitals,
                                               self.n_alpha,
                                               self.n_beta)
        qubit_op = fermion_to_qubit_mapping(fermion_operator=fermion_op,
                                            mapping=self.mapping,
                                            n_spinorbitals=self.n_spinorbitals,
                                            n_electrons=self.n_electrons,
                                            up_then_down=self.up_then_down)

        # Cast all coefs to floats (rotations angles are real)
        for key in qubit_op.terms:
            qubit_op.terms[key] = float(qubit_op.terms[key].imag)
        qubit_op.compress()
        return qubit_op

    def _compute_mp2_params(self):
        """Computes the MP2 initial variational parameters. Compute the initial
        variational parameters with PySCF MP2 calculation, and then reorders the
        elements into the appropriate convention. MP2 only has doubles (T2)
        amplitudes, thus the single (T1) amplitudes are set to a small non-zero
        value and added. The ordering is single, double (diagonal), double
        (non-diagonal).

        Returns:
            list of float: The initial variational parameters.
        """

        mp2_fragment = mp.MP2(self.molecule.mean_field, frozen=self.molecule.frozen_mos)
        mp2_fragment.verbose = 0
        _, mp2_t2 = mp2_fragment.kernel()

        # Get singles amplitude. Just get "up" amplitude, since "down" should be the same
        singles = [2.e-5] * (self.n_virtual * self.n_occupied)

        # Get singles and doubles amplitudes associated with one spatial occupied-virtual pair
        doubles_1 = [-mp2_t2[q, q, p, p]/2. if (abs(-mp2_t2[q, q, p, p]/2.) > 1e-15) else 0.
                     for p, q in itertools.product(range(self.n_virtual), range(self.n_occupied))]

        # Get doubles amplitudes associated with two spatial occupied-virtual pairs
        doubles_2 = [-mp2_t2[q, s, p, r] for (p, q), (r, s)
                     in itertools.combinations(itertools.product(range(self.n_virtual), range(self.n_occupied)), 2)]

        return singles + doubles_1 + doubles_2



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import numpy as np

from tangelo.linq import Gate, Circuit
from tangelo.toolboxes.ansatz_generator.variational_circuit import VariationalCircuitAnsatz


# UCC1 hard coding circuit. Simple example not relying on import Ansatz.
lst_gates = [Gate("RX", 0, parameter=np.pi/2)]
lst_gates += [Gate("H", qubit_i) for qubit_i in range(1, 4)]
lst_gates += [Gate("CNOT", qubit_i+1, qubit_i) for qubit_i in range(3)]
lst_gates += [Gate("RZ", 3, parameter="theta", is_variational=True)]
lst_gates += [Gate("CNOT", qubit_i, qubit_i-1) for qubit_i in range(3, 0, -1)]
lst_gates += [Gate("H", qubit_i) for qubit_i in range(3, 0, -1)]
lst_gates += [Gate("RX", 0, parameter=-np.pi/2)]
circuit = Circuit(lst_gates)


class VariationalCircuitTest(unittest.TestCase):

    def test_init(self):
        """Test initialization of the ansatz class."""
        VariationalCircuitAnsatz(circuit)

    def test_set_var_params(self):
        """Test setting variational parameters."""
        circuit_ansatz = VariationalCircuitAnsatz(circuit)

        single_ones = np.array([1.])

        circuit_ansatz.set_var_params("ones")
        np.testing.assert_array_almost_equal(circuit_ansatz.var_params,  single_ones, decimal=6)

        circuit_ansatz.set_var_params([1.])
        np.testing.assert_array_almost_equal(circuit_ansatz.var_params,  single_ones, decimal=6)

        circuit_ansatz.set_var_params(np.array([1.]))
        np.testing.assert_array_almost_equal(circuit_ansatz.var_params,  single_ones, decimal=6)

    def test_uccsd_incorrect_number_var_params(self):
        """Returns an error if user provide incorrect number of variational parameters."""

        circuit_ansatz = VariationalCircuitAnsatz(circuit)
        self.assertRaises(ValueError, circuit_ansatz.set_var_params, np.array([1., 1.]))


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic noise model representation and backend-specific translation function.
The Simulator class is responsible for taking in the generic noise model at
runtime and applying it accordingly to the target compute backend.

Only works for simulators supporting noisy simulation.
"""


from tangelo.linq import ONE_QUBIT_GATES

SUPPORTED_NOISE_MODELS = {'depol', 'pauli'}

__MAPPING_GATES_QISKIT = dict()
for name in ONE_QUBIT_GATES:
    __MAPPING_GATES_QISKIT[name] = [name.lower()]
for name in {'RX', 'RY', 'RZ'}:
    __MAPPING_GATES_QISKIT[name] = ['u1', 'u2', 'u3']
__MAPPING_GATES_QISKIT["CNOT"] = ["cx"]


class NoiseModel:
    """A class representing a noise model. Not specific to any compute backend.
    The object holds a dictionary mapping each noisy gate to a list of noise
    that should be applied every time it is encountered in a circuit.

    Pauli noise expects a list of 3 probabilities corresponding to X,Y,Z pauli
    noises (I is deduced from the sum) Depolarization noise is a special case of
    Pauli noise, where the 3 probabilities are equal: a float is enough. Please
    check the notebook tutorial in the example section more a more in-depth
    description of the different noise channels.
    """

    def __init__(self):
        self._quantum_errors = dict()

    def __repr__(self):
        return str(self._quantum_errors)

    def add_quantum_error(self, abs_gate, noise_type, noise_params):
        """Adds the desired noise to the gate specified by user. Checks if
        inputs makes sense.
        """

        if noise_type not in SUPPORTED_NOISE_MODELS:
            raise ValueError(f"Error model not supported :{noise_type}\nCurrently supported:{SUPPORTED_NOISE_MODELS}")
        if noise_type == 'pauli' and (not isinstance(noise_params, list) or len(noise_params) != 3):
            raise ValueError(f"For pauli noise, a list of 3 probabilities (corresponding to X,Y,Z) is expected")
        if noise_type == 'depol' and not isinstance(noise_params, float):
            raise ValueError(f"For depolarization noise, the expected parameter must be a single float")

        if abs_gate in self._quantum_errors:
            if noise_type not in {nt for nt, np in self._quantum_errors[abs_gate]}:
                self._quantum_errors[abs_gate] += [(noise_type, noise_params)]
            else:
                raise ValueError(f'Cannot cumulate several noise channels of the same type on the same gate')
        else:
            self._quantum_errors[abs_gate] = [(noise_type, noise_params)]

    @property
    def noisy_gates(self):
        return set(self._quantum_errors.keys())


def get_qiskit_noise_dict(noise_model):
    """Takes in generic noise model, converts to a dictionary whose keys are
    the qiskit basis gates supported by the qasm simulator, to the noises,
    ensuring there are no redundancy / duplicates on the U-gates in particular.
    """

    qnd = dict()
    for gate, noises in noise_model._quantum_errors.items():
        for qiskit_gate in __MAPPING_GATES_QISKIT[gate]:
            if qiskit_gate not in qnd:
                qnd[qiskit_gate] = noises
            else:
                noise_types = [nt for nt, np in qnd[qiskit_gate]]
                for noise in noises:
                    if noise[0] not in noise_types:
                        noise_types.append(noise[0])
                        qnd[qiskit_gate].append(noise)
    return qnd


def get_qiskit_noise_model(noise_model):
    """Takes a NoiseModel object as input, returns an equivalent Qiskit Noise
    Model, compatible with QASM simulator.
    """

    from qiskit.providers.aer.noise import NoiseModel as QiskitNoiseModel
    from qiskit.providers.aer.noise.errors import depolarizing_error, pauli_error

    qnd = get_qiskit_noise_dict(noise_model)
    qnm = QiskitNoiseModel()
    for gate, noises in qnd.items():
        for noise, probs in noises:
            if noise == 'pauli':
                qerr1 = pauli_error(list(zip(['X', 'Y', 'Z'], probs)) + [('I', 1-sum(probs))])
                qerr2 = qerr1.tensor(qerr1)
                qerr = qerr1 if gate in qnm._1qubit_instructions else qerr2
            elif noise == 'depol':
                qerr = depolarizing_error(probs, 1 if gate in qnm._1qubit_instructions else 2)
            else:
                raise ValueError(f"Error model currently not supported with Qiskit backend :{noise}")
            qnm.add_all_qubit_quantum_error(qerr, gate)
    return qnm



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from tangelo.molecule_library import mol_H2_sto3g
from tangelo.toolboxes.qubit_mappings import jordan_wigner


def assert_term_dict_almost_equal(d1, d2, delta=1e-10):
    """Utility function to check whether two qubit operators are almost equal,
    looking at their term dictionary, for an arbitrary absolute tolerance.
    """
    d1k, d2k = set(d1.keys()), set(d2.keys())
    if d1k != d2k:
        d1_minus_d2 = d1k.difference(d2k)
        d2_minus_d1 = d2k.difference(d1k)
        raise AssertionError("Term dictionary keys differ. Qubit operators are not almost equal.\n"
                             f"d1-d2 keys: {d1_minus_d2} \nd2-d1 keys: {d2_minus_d1}")
    else:
        for k in d1k:
            if abs(d1[k] - d2[k]) > delta:
                raise AssertionError(f"Term {k}, difference={abs(d1[k]-d2[k])} > delta={delta}:\n {d1[k]} != {d2[k]}")


class QubitizerTest(unittest.TestCase):

    def test_qubit_hamiltonian_JW_h2(self):
        """Verify computation of the Jordan-Wigner Hamiltonian for the H2
        molecule."""

        qubit_hamiltonian = jordan_wigner(mol_H2_sto3g.fermionic_hamiltonian)

        # Obtained with Openfermion
        reference_terms = {(): -0.0988639693354571, ((0, "Z"),): 0.17119774903432955, ((1, "Z"),): 0.17119774903432958,
                           ((2, "Z"),): -0.22278593040418496, ((3, "Z"),): -0.22278593040418496,
                           ((0, "Z"), (1, "Z")): 0.16862219158920938, ((0, "Z"), (2, "Z")): 0.120544822053018,
                           ((0, "Z"), (3, "Z")): 0.165867024105892, ((1, "Z"), (2, "Z")): 0.165867024105892,
                           ((1, "Z"), (3, "Z")): 0.120544822053018, ((2, "Z"), (3, "Z")): 0.17434844185575687,
                           ((0, "X"), (1, "X"), (2, "Y"), (3, "Y")): -0.045322202052874,
                           ((0, "X"), (1, "Y"), (2, "Y"), (3, "X")): 0.045322202052874,
                           ((0, "Y"), (1, "X"), (2, "X"), (3, "Y")): 0.045322202052874,
                           ((0, "Y"), (1, "Y"), (2, "X"), (3, "X")): -0.045322202052874}

        assert_term_dict_almost_equal(qubit_hamiltonian.terms, reference_terms, delta=1e-8)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bath orbital construction for DMET calculation.

The construction of the bath orbitals (the orbitals which include the
environment effect from the surrounding part) is done here.
"""

import numpy as np


def dmet_fragment_bath(mol, t_list, temp_list, onerdm_low):
    """ Construct the bath orbitals for DMET fragment calculation.

    Args:
        mol (pyscf.gto.Mole): The molecule to simulate (The full molecular
            system).
        t_list (list): Number of [0] fragment & [1] bath orbitals (int).
        temp_list (list): [0] Minimum and [1] maximum number for the active
            orbitals (int).
        onerdm_low (numpy.array): One-particle RDM from the low-level
            calculation (float64).

    Returns:
        numpy.array: The bath orbitals (float64).
        numpy.array: Orbital energies (float64).
    """

    # Extract the one-particle RDM for the active space
    onerdm_embedded = dmet_onerdm_embed(mol, temp_list, onerdm_low)

    # Diagonalize it
    e, c = np.linalg.eigh(onerdm_embedded)

    # Sort the eigenvectors with the eigenvalues
    e_sorted, c_sorted = dmet_bath_orb_sort(t_list, e, c)

    # Add the core contribution
    bath_orb, e_core = dmet_add_to_bath_orb(mol, t_list, temp_list, e_sorted, c_sorted)

    return bath_orb, e_core


def dmet_onerdm_embed(mol, temp_list, onerdm_before):
    """ Extract the one particle RDM of the active space.

    Args:
        mol (pyscf.gto.Mole): The molecule to simulate (The full molecular
            system).
        temp_list (list): [0] Minimum and [1] maximum number for the active
            orbitals (int).
        onerdm_before (numpy.array): One-particle RDM from the low-level
            calculation (float64).

    Returns:
        numpy.array: Extracted one-particle RDM (float64).
    """

    # Get the number of orbitals
    norbital_total = mol.nao_nr()

    # Reshape the RDM
    onerdm_matrix = np.reshape(onerdm_before, (norbital_total, norbital_total))

    if temp_list[0] == 0:
        # If it is the first fragment, just determine the maximum for extraction
        onerdm_temp = onerdm_matrix[:, temp_list[1]:]
        onerdm_temp3 = onerdm_temp[temp_list[1]:, :]
    else:
        # Determine the minimum and maximum orbitals for extraction
        onerdm_temp = onerdm_matrix[:, : temp_list[0]]
        onerdm_temp2 = onerdm_matrix[:, temp_list[1]:]
        onerdm_temp3 = np.hstack((onerdm_temp, onerdm_temp2))
        onerdm_temp = onerdm_temp3[:temp_list[0], :]
        onerdm_temp2 = onerdm_temp3[temp_list[1]:, :]
        onerdm_temp3 = np.vstack((onerdm_temp, onerdm_temp2))

    return onerdm_temp3


def dmet_bath_orb_sort(t_list, e_before, c_before):
    """ Sort the bath orbitals with the eigenvalues (orbital energies).

    Args:
        t_list (list): Number of [0] fragment & [1] bath orbitals (int).
        e_before (numpy.array): Orbitals energies before sorting (float64).
        c_before (numpy.array): Coefficients of the orbitals before sorting
            (float64).

    Returns:
        numpy.array: Sorted orbital energies (float64).
        numpy.array: Coefficients of the sorted orbitals (float64).
    """

    # Sort the orbital energies (Occupation of 1.0 should come first...)
    new_index = np.maximum(-e_before, e_before - 2.0).argsort()

    # Throw away some orbitals above threshold
    thresh_orb = np.sum(-np.maximum(-e_before, e_before - 2.0)[new_index] > 1e-13)

    # Determine the number of bath orbitals
    norb = min(np.sum(thresh_orb), t_list[0])
    t_list.append(norb)

    # Sort the bath orbitals with its energies
    e_new = e_before[new_index]
    c_new = c_before[:, new_index]

    return e_new, c_new


def dmet_add_to_bath_orb(mol, t_list, temp_list, e_before, c_before):
    """ Add the frozen core part to the bath orbitals.

    Args:
        mol (pyscf.gto.Mole): The molecule to simulate (The full molecular
            system).
        t_list (list): Number of [0] fragment & [1] bath orbitals (int).
        temp_list (list): [0] Minimum and [1] maximum number for the active
            orbitals (int).
        e_before (numpy.array): Orbital energy before addition of frozen core
            (float64).
        c_before (numpy.array): Coefficients of the orbitals before addition of
            frozen core (float64).

    Returns:
        numpy.array: Constructed bath orbitals (float64).
        numpy.array: Orbital energies (float64).
    """

    # Copy the bath orbitals and energies be fore adding the core
    add_e = - e_before[t_list[1]:]
    add_c = c_before[:, t_list[1]:]
    new_index = add_e.argsort()

    # Sort the orbitals based on its energies
    c_before[:, t_list[1]:] = add_c[:, new_index]
    add_e = - add_e[new_index]

    # The orbital energies with core part
    norbital_total = mol.nao_nr()
    e_occupied_core_orbitals = np.hstack((np.zeros([t_list[0] + t_list[1]]), add_e))

    # Add the core part in the orbitals
    for orb in range(0, t_list[0]):
        c_before = np.insert(c_before, orb, 0.0, axis=1)
    i_temp = 0
    for orb_total in range(0, norbital_total):
        if ((orb_total >= temp_list[0]) and (orb_total < temp_list[1])):
            c_before = np.insert(c_before, orb_total, 0.0, axis=0)
            c_before[orb_total, i_temp] = 1.0
            i_temp += 1

    return c_before, e_occupied_core_orbitals



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines the penatly terms that can be added to the target
fermionic Hamiltonian, providing the ability to restrict the Hilbert space of
solutions using VQE. For example usages see
    - Illa G. Ryabinkin, Scott N. Genin, Artur F. Izmaylov. "Constrained
        variational quantum eigensolver: Quantum computer search engine in the
        Fock space" https://arxiv.org/abs/1806.00461.
    - Gabriel Greene-Diniz, David Munoz Ramo. "Generalized unitary coupled
        cluster excitations for multireference molecular states optimized by the
        Variational Quantum Eigensolver" https://arxiv.org/abs/1910.05168.
"""

from tangelo.toolboxes.ansatz_generator.fermionic_operators import number_operator_list, spinz_operator_list, spin2_operator_list
from tangelo.toolboxes.operators import FermionOperator, squared_normal_ordered


def number_operator_penalty(n_orbs, n_electrons, mu=1, up_then_down=False):
    r"""Function to generate the normal ordered number operator penalty term as
    a FermionOperator.

    Args:
        n_orbs (int): number of orbitals in the fermion basis (this is number of
            spin-orbitals divided by 2).
        n_electrons (int): number of electrons.
        mu (float): Positive number in front of penalty term.
        up_then_down (bool): The ordering of the spin orbitals.
            qiskit (True) openfermion (False)
            If later transforming to qubits, one should generally let the qubit
            mapping handle the ordering.

    Returns:
        FermionOperator: The number operator penalty term
            mu*(\hat{N}-n_electrons)^2.
    """

    all_terms = [[(), -n_electrons]] + number_operator_list(n_orbs, up_then_down)

    return mu*squared_normal_ordered(all_terms)


def spin_operator_penalty(n_orbs, sz, mu=1, up_then_down=False):
    r"""Function to generate the normal ordered Sz operator penalty term.

    Args:
        n_orbs (int): number of orbitals in the fermion basis (this is number of
            spin-orbitals divided by 2).
        sz (int): the desired Sz quantum number to penalize for.
        mu (float): Positive number in front of penalty term.
        up_then_down: The ordering of the spin orbitals.
            qiskit (True) openfermion (False)
            If later transforming to qubits, one should generally let the qubit
            mapping handle the ordering.

    Returns:
        FermionOperator: The Sz operator penalty term mu*(\hat{Sz}-sz)^2.
    """

    all_terms = [[(), -sz]] + spinz_operator_list(n_orbs, up_then_down)

    return mu*squared_normal_ordered(all_terms)


def spin2_operator_penalty(n_orbs, s2, mu=1, up_then_down=False):
    r"""Function to generate the normal ordered S^2 operator penalty term,
    operator form taken from
    https://pubs.rsc.org/en/content/articlepdf/2019/cp/c9cp02546d.

    Args:
        n_orbs (int): number of orbitals in the fermion basis (this is number of
            spin-orbitals divided by 2).
        s2 (int): the desired S^2 quantum number to penalize for.
            singlet: s2=0*(0+1)=0, doublet: s2=(1/2)*(1/2+1)=3/4, triplet,
            s2=1*(1+1)=2 ...
        mu (float): Positive number in front of penalty term.
        up_then_down: The ordering of the spin orbitals.
            qiskit (True) openfermion (False)
            If later transforming to qubits, one should generally let the qubit
            mapping handle the ordering.

    Returns:
        FermionOperator: The S^2 operator penalty term mu*(\hat{S}^2-s2)^2.
    """

    all_terms = [[(), -s2]] + spin2_operator_list(n_orbs, up_then_down)

    return mu*squared_normal_ordered(all_terms)


def combined_penalty(n_orbs, opt_penalty_terms=None, up_then_down=False):
    r"""Function to generate the sum of all available penalty terms, currently
    implemented are
    - "N": number operator with eigenvalue (number of particles).
    - "Sz": Sz|s,m_s> = ms|s,m_s>.
    - "S^2": S^2|s,m_s> = s(s+1)|s,m_s>.

    Args:
        n_orbs (int): number of active orbitals in the fermion basis (this is
            number of spin-orbitals divided by 2).
        opt_penalty_terms (dict): The options for each penalty "N", "Sz", "S^2"
            as
            - "N" (array or list[float]):
                [Prefactor, Value] Prefactor * (\hat{N} - Value)^2
            - "Sz" (array or list[float]):
                [Prefactor, Value] Prefactor * (\hat{Sz} - Value)^2
            - "S^2" (array or list[float]):
                [Prefactor, Value] Prefactor * (\hat{S}^2 - Value)^2
        up_then_down: The ordering of the spin orbitals.
            qiskit (True) openfermion (False)
            If later transforming to qubits, one should generally let the qubit
            mapping handle this.

    Returns:
        FermionOperator: The combined n_electron+sz+s^2 penalty
            terms.
    """

    penalty_terms = {"N": [0, 0], "Sz": [0, 0], "S^2": [0, 0]}
    if opt_penalty_terms:
        for k, v in opt_penalty_terms.items():
            if k in penalty_terms:
                penalty_terms[k] = v
            else:
                raise KeyError(f"Keyword :: {k}, penalty term not available")
    else:
        return FermionOperator()

    pen_ferm = FermionOperator()
    if (penalty_terms["N"][0] > 0):
        prefactor, n_electrons = penalty_terms["N"][:]
        pen_ferm += number_operator_penalty(n_orbs, n_electrons, mu=prefactor, up_then_down=up_then_down)
    if (penalty_terms["Sz"][0] > 0):
        prefactor, sz = penalty_terms["Sz"][:]
        pen_ferm += spin_operator_penalty(n_orbs, sz, mu=prefactor, up_then_down=up_then_down)
    if (penalty_terms["S^2"][0] > 0):
        prefactor, s2 = penalty_terms["S^2"][:]
        pen_ferm += spin2_operator_penalty(n_orbs, s2, mu=prefactor, up_then_down=up_then_down)
    return pen_ferm



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions helping with quantum circuit format conversion between abstract
format and qulacs format.

In order to produce an equivalent circuit for the target backend, it is
necessary to account for:
- how the gate names differ between the source backend to the target backend.
- how the order and conventions for some of the inputs to the gate operations
    may also differ.
"""
from numpy import exp, cos, sin


def get_qulacs_gates():
    """Map gate name of the abstract format to the equivalent add_gate method of
    Qulacs's QuantumCircuit class API and supported gates:
    http://qulacs.org/class_quantum_circuit.html
    """
    import qulacs

    GATE_QULACS = dict()
    GATE_QULACS["H"] = qulacs.QuantumCircuit.add_H_gate
    GATE_QULACS["X"] = qulacs.QuantumCircuit.add_X_gate
    GATE_QULACS["Y"] = qulacs.QuantumCircuit.add_Y_gate
    GATE_QULACS["Z"] = qulacs.QuantumCircuit.add_Z_gate
    GATE_QULACS["CH"] = qulacs.gate.H
    GATE_QULACS["CX"] = qulacs.gate.X
    GATE_QULACS["CY"] = qulacs.gate.Y
    GATE_QULACS["CZ"] = qulacs.gate.Z
    GATE_QULACS["S"] = qulacs.QuantumCircuit.add_S_gate
    GATE_QULACS["T"] = qulacs.QuantumCircuit.add_T_gate
    GATE_QULACS["RX"] = qulacs.QuantumCircuit.add_RX_gate
    GATE_QULACS["RY"] = qulacs.QuantumCircuit.add_RY_gate
    GATE_QULACS["RZ"] = qulacs.QuantumCircuit.add_RZ_gate
    GATE_QULACS["CNOT"] = qulacs.QuantumCircuit.add_CNOT_gate
    GATE_QULACS["CRX"] = qulacs.gate.RX
    GATE_QULACS["CRY"] = qulacs.gate.RY
    GATE_QULACS["CRZ"] = qulacs.gate.RZ
    GATE_QULACS["PHASE"] = qulacs.gate.DenseMatrix
    GATE_QULACS["CPHASE"] = qulacs.gate.DenseMatrix
    GATE_QULACS["XX"] = qulacs.gate.DenseMatrix
    GATE_QULACS["SWAP"] = qulacs.QuantumCircuit.add_SWAP_gate
    GATE_QULACS["CSWAP"] = qulacs.gate.SWAP
    GATE_QULACS["MEASURE"] = qulacs.gate.Measurement
    return GATE_QULACS


def translate_qulacs(source_circuit, noise_model=None):
    """Take in an abstract circuit, return an equivalent qulacs QuantumCircuit
    instance. If provided with a noise model, will add noisy gates at
    translation. Not very useful to look at, as qulacs does not provide much
    information about the noisy gates added when printing the "noisy circuit".

    Args:
        source_circuit: quantum circuit in the abstract format.
        noise_model: A NoiseModel object from this package, located in the
            noisy_simulation subpackage.

    Returns:
        qulacs.QuantumCircuit: the corresponding qulacs quantum circuit.
    """

    import qulacs
    from qulacs.gate import X, Y, Z, Probabilistic, DepolarizingNoise, TwoQubitDepolarizingNoise

    GATE_QULACS = get_qulacs_gates()
    target_circuit = qulacs.QuantumCircuit(source_circuit.width)

    # Maps the gate information properly. Different for each backend (order, values)
    for gate in source_circuit._gates:
        if gate.name in {"H", "X", "Y", "Z", "S", "T"}:
            (GATE_QULACS[gate.name])(target_circuit, gate.target[0])
        elif gate.name in {"CH", "CX", "CY", "CZ"}:
            mat_gate = qulacs.gate.to_matrix_gate(GATE_QULACS[gate.name](gate.target[0]))
            for c in gate.control:
                mat_gate.add_control_qubit(c, 1)
            target_circuit.add_gate(mat_gate)
        elif gate.name in {"RX", "RY", "RZ"}:
            (GATE_QULACS[gate.name])(target_circuit, gate.target[0], -1. * gate.parameter)
        elif gate.name in {"CRX", "CRY", "CRZ"}:
            mat_gate = qulacs.gate.to_matrix_gate(GATE_QULACS[gate.name](gate.target[0], -1. * gate.parameter))
            for c in gate.control:
                mat_gate.add_control_qubit(c, 1)
            target_circuit.add_gate(mat_gate)
        elif gate.name in {"SWAP"}:
            (GATE_QULACS[gate.name])(target_circuit, gate.target[0], gate.target[1])
        elif gate.name in {"CSWAP"}:
            mat_gate = qulacs.gate.to_matrix_gate(GATE_QULACS[gate.name](gate.target[0], gate.target[1]))
            for c in gate.control:
                mat_gate.add_control_qubit(c, 1)
            target_circuit.add_gate(mat_gate)
        elif gate.name in {"PHASE"}:
            mat_gate = GATE_QULACS[gate.name](gate.target[0], [[1, 0], [0, exp(1j * gate.parameter)]])
            target_circuit.add_gate(mat_gate)
        elif gate.name in {"CPHASE"}:
            mat_gate = GATE_QULACS[gate.name](gate.target[0], [[1, 0], [0, exp(1j * gate.parameter)]])
            for c in gate.control:
                mat_gate.add_control_qubit(c, 1)
            target_circuit.add_gate(mat_gate)
        elif gate.name in {"XX"}:
            c = cos(gate.parameter/2)
            s = -1j * sin(gate.parameter/2)
            mat_gate = GATE_QULACS[gate.name]([gate.target[0], gate.target[1]], [[c, 0, 0, s],
                                                                                 [0, c, s, 0],
                                                                                 [0, s, c, 0],
                                                                                 [s, 0, 0, c]])
            target_circuit.add_gate(mat_gate)
        elif gate.name in {"CNOT"}:
            (GATE_QULACS[gate.name])(target_circuit, gate.control[0], gate.target[0])
        elif gate.name in {"MEASURE"}:
            gate = (GATE_QULACS[gate.name])(gate.target[0], gate.target[0])
            target_circuit.add_gate(gate)
        else:
            raise ValueError(f"Gate '{gate.name}' not supported on backend qulacs")

        # Add noisy gates
        if noise_model and (gate.name in noise_model.noisy_gates):
            for nt, np in noise_model._quantum_errors[gate.name]:
                if nt == 'pauli':
                    for t in gate.target:
                        target_circuit.add_gate(Probabilistic(np, [X(t), Y(t), Z(t)]))
                    if gate.control is not None:
                        for c in gate.control:
                            target_circuit.add_gate(Probabilistic(np, [X(c), Y(c), Z(c)]))
                elif nt == 'depol':
                    depol_list = [t for t in gate.target]
                    if gate.control is not None:
                        depol_list += [c for c in gate.control]
                    n_depol = len(depol_list)
                    if n_depol == 2:
                        target_circuit.add_gate(TwoQubitDepolarizingNoise(*depol_list, (15/16)*np))
                    elif n_depol == 1:
                        target_circuit.add_gate(DepolarizingNoise(depol_list[0], (3/4) * np))
                    else:
                        raise ValueError(f'{gate.name} has more than 2 qubits, Qulacs DepolarizingNoise only supports 1- and 2-qubits')

    return target_circuit



import setuptools
import sys
import subprocess


def install(package):
    subprocess.check_call([sys.executable, "-m", "pip", "install", package])

with open("tangelo/_version.py") as f:
    version = f.readlines()[-1].split()[-1].strip("\"'")

with open('README.rst', 'r') as f:
    long_description = f.read()

install('wheel')
install('pyscf')
install('git+https://github.com/pyscf/semiempirical')

description = "Maintained by Good Chemistry Company, focusing on the development of end-to-end materials simulation workflows on quantum computers."

setuptools.setup(
    name="tangelo-gc",
    author="The Tangelo developers",
    version=version,
    description=description,
    long_description=description,
    #long_description_content_type=description,
    url="https://github.com/goodchemistryco/Tangelo",
    packages=setuptools.find_packages(),
    test_suite="tangelo",
    setup_requires=['h5py'],
    install_requires=['h5py', 'bitarray', 'openfermion', 'openfermionpyscf']
)



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Construct localized orbitals for DMET calculation

The construction of the localized orbitals from the orbitals obtained from the
full system mean-field calculation is done here.

Localization schemes based on Intrisic atomic orbitals (IAO) (dmet_iao.py),
Lowdin, meta-Lowdin, and Boys can be selected. The latter three rely on external
PySCF program. However, using Boys localization is not recommended.
"""

from pyscf import scf, ao2mo
import numpy as np
from functools import reduce


class dmet_orbitals:
    """ Localize the SCF orbitals and calculate the integrals

    This class handles the information from the calculation of the entire
    molecular system

    Attributes:
        mol_full (pyscf.gto.Mole): The molecule to simulate (The full molecule).
        mf_full (pyscf.scf.RHF): The mean-field of the molecule (The full
            molecule).
        low_scf_energy (float64): The mean-field total energy of the full
            molecule.
        low_scf_fock (numpy.array): The Fock matrix for the full molecule
            (float64).
        dmet_active_orbitals (list): Label for active orbitals. "1" is active,
            and "0" is inactive (int).
        number_active_orbitals (int): Number of active orbitals.
        number_active_electrons (int): Number of active electrons.
        localized_mo (numpy.array): The localized molecular orbitals (float64).
        core_constant_energy (float64): Constant core energy (Constant value
            which does not change in DMET).
        active_oneint (numpy.array): One-electron integral in localized MO basis
            (float64).
        active_fock (numpy.array): Fock matrix in localized MO basis (float64).
    """

    def __init__(self, mol, mf, active_space, localization_function):
        """Initialize the class.

        Localize the orbitals, get energies and integrals for the entire system
        in localized MO basis.

        Args:
            mol (pyscf.gto.Mole): The molecule to simulate (The full molecule).
            mf (pyscf.scf.RHF): The mean field of the molecule (The full
                molecule).
            active_space (list): The active space in DMET calculation. All
                orbitals in the initial SCF calculation (int).
            localization_function (string): Localization scheme.
        """

        # TODO: Is active space always claculated from the molecule?

        # Obtain the elements from the low-level SCF calculations
        self.mol_full = mol
        self.mf_full = mf
        self.low_scf_energy = mf.e_tot
        low_scf_dm = reduce(np.dot, (mf.mo_coeff, np.diag(mf.mo_occ), mf.mo_coeff.T))
        low_scf_twoint = scf.hf.get_veff(mf.mol, low_scf_dm, 0, 0, 1)
        self.low_scf_fock = mf.mol.intor("cint1e_kin_sph") + mf.mol.intor("cint1e_nuc_sph") + low_scf_twoint

        # Define the active space if possible
        self.dmet_active_orbitals = np.zeros([mf.mol.nao_nr()], dtype=int)
        self.dmet_active_orbitals[active_space] = 1
        self.number_active_orbitals = np.sum(self.dmet_active_orbitals)
        self.number_active_electrons = int(np.rint(mf.mol.nelectron - np.sum(mf.mo_occ[self.dmet_active_orbitals == 0])))

        # Localize the orbitals (IAO)
        self.localized_mo = localization_function(mol, mf)

        # Define the core space if possible (Initial calculations treat the entire molecule ...)
        core_mo_dm = np.array(mf.mo_occ, copy=True)
        core_mo_dm[self.dmet_active_orbitals == 1] = 0
        core_ao_dm = reduce(np.dot, (mf.mo_coeff, np.diag(core_mo_dm), mf.mo_coeff.T))
        core_twoint = scf.hf.get_veff(mf.mol, core_ao_dm, 0, 0, 1)
        core_oneint = self.low_scf_fock - low_scf_twoint + core_twoint

        # Define the energies and matrix elements based on the localized orbitals
        self.core_constant_energy = mf.mol.energy_nuc() + np.einsum("ij,ij->", core_oneint - 0.5*core_twoint, core_ao_dm)
        self.active_oneint = reduce(np.dot, (self.localized_mo.T, core_oneint, self.localized_mo))
        self.active_fock = reduce(np.dot, (self.localized_mo.T, self.low_scf_fock, self.localized_mo))

    def dmet_fragment_hamiltonian(self, bath_orb, norb_high, onerdm_core):
        """Construct the Hamiltonian for a DMET fragment.

        Args:
            bath_orb (numpy.array): The bath orbitals (float64).
            norb_high (int): Number of orbitals in the fragment calculation.
            onerdm_core (numpy.array): The core part of the one-particle RDM
                (float64).

        Returns:
            numpy.array: One-electron integrals for fragment calculation
                (float64).
            numpy.array: The fock matrix for fragment calculation (float64).
            numpy.array: Two-electron integrals for fragment calculation
                (float64).
        """

        # Calculate one-electron integrals
        frag_oneint = reduce(np.dot, (bath_orb[:, : norb_high].T, self.active_oneint, bath_orb[:, : norb_high]))

        # Calculate the fock matrix
        density_matrix = reduce(np.dot, (self.localized_mo, onerdm_core, self.localized_mo.T))
        two_int = scf.hf.get_veff(self.mol_full, density_matrix, 0, 0, 1)
        new_fock = self.active_oneint + reduce(np.dot, ((self.localized_mo.T, two_int, self.localized_mo)))
        frag_fock = reduce(np.dot, (bath_orb[:, : norb_high].T, new_fock, bath_orb[:, : norb_high]))

        # Calculate the two-electron integrals
        coefficients = np.dot(self.localized_mo, bath_orb[:, : norb_high])
        frag_twoint = ao2mo.outcore.full_iofree(self.mol_full, coefficients, compact=False).reshape(
                                                norb_high,  norb_high,  norb_high,  norb_high)

        return frag_oneint, frag_fock, frag_twoint



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from tangelo.algorithms.variational import ADAPTSolver
from tangelo.molecule_library import mol_H2_sto3g, xyz_H4
from tangelo.toolboxes.ansatz_generator._unitary_majorana_cc import get_majorana_uccgsd_pool, get_majorana_uccsd_pool
from tangelo.toolboxes.molecular_computation.molecule import SecondQuantizedMolecule


class ADAPTSolverTest(unittest.TestCase):

    def test_build_adapt(self):
        """Try instantiating ADAPTSolver with basic input."""

        opt_dict = {"molecule": mol_H2_sto3g, "max_cycles": 15}
        adapt_solver = ADAPTSolver(opt_dict)
        adapt_solver.build()

    def test_single_cycle_adapt(self):
        """Try instantiating ADAPTSolver with basic input. The fermionic term
        ordering has been taken from the reference below (original paper for
        ADAPT-VQE).

        Reference:
            - Grimsley, H.R., Economou, S.E., Barnes, E. et al.
            An adaptive variational algorithm for exact molecular simulations on
            a quantum computer. Nat Commun 10, 3007 (2019).
            https://doi.org/10.1038/s41467-019-10988-2
        """

        opt_dict = {"molecule": mol_H2_sto3g, "max_cycles": 1, "verbose": False}
        adapt_solver = ADAPTSolver(opt_dict)
        adapt_solver.build()
        adapt_solver.simulate()

        self.assertAlmostEqual(adapt_solver.optimal_energy, -1.13727, places=3)

        resources = {"qubit_hamiltonian_terms": 15,
                     "circuit_width": 4,
                     "circuit_gates": 122,
                     "circuit_2qubit_gates": 48,
                     "circuit_var_gates": 8,
                     "vqe_variational_parameters": 1}
        self.assertEqual(adapt_solver.get_resources(), resources)

    def test_multiple_cycle_adapt_majorana_pool(self):
        """Solve H4 with one frozen orbtial with ADAPTSolver using 4 cycles and operators chosen
        from a Majorana UCCGSD pool and a Majorana UCCSD pool
        """

        mol = SecondQuantizedMolecule(xyz_H4, 0, 0, "sto-3g", frozen_orbitals=[0])
        opt_dict = {"molecule": mol, "max_cycles": 4, "verbose": False, "pool": get_majorana_uccgsd_pool,
                    "pool_args": {"n_sos": mol.n_active_sos}}
        adapt_solver = ADAPTSolver(opt_dict)
        adapt_solver.build()
        adapt_solver.simulate()

        self.assertAlmostEqual(adapt_solver.optimal_energy, -1.8945, places=3)

        mol = SecondQuantizedMolecule(xyz_H4, 0, 0, "sto-3g", frozen_orbitals=[0])
        opt_dict = {"molecule": mol, "max_cycles": 4, "verbose": False, "pool": get_majorana_uccsd_pool,
                    "pool_args": {"n_electrons": mol.n_active_electrons, "n_sos": mol.n_active_sos}}
        adapt_solver = ADAPTSolver(opt_dict)
        adapt_solver.build()
        adapt_solver.simulate()

        self.assertAlmostEqual(adapt_solver.optimal_energy, -1.8945, places=3)

        deflation_circuits = [adapt_solver.optimal_circuit]
        ref_state = [0, 1, 0, 0, 0, 1]
        opt_dict = {"molecule": mol, "max_cycles": 1, "verbose": False, "pool": get_majorana_uccgsd_pool,
                    "pool_args": {"n_sos": mol.n_active_sos},
                    "deflation_circuits": deflation_circuits, "ref_state": ref_state, "deflation_coeff": 1}
        adapt_solver = ADAPTSolver(opt_dict)
        adapt_solver.build()
        optimal_energy = adapt_solver.simulate()

        self.assertAlmostEqual(optimal_energy, -1.91062, places=3)

    def test_multiple_cycle_adapt_majorana_pool_with_deflation(self):
        """Solve H4 with one frozen orbtial with ADAPTSolver using 4 cycles and operators chosen
        from a Majorana UCCGSD pool followed by deflation for an orthogonal state triplet state.
        """

        mol = SecondQuantizedMolecule(xyz_H4, 0, 0, "sto-3g", frozen_orbitals=[0])
        opt_dict = {"molecule": mol, "max_cycles": 4, "verbose": False, "pool": get_majorana_uccgsd_pool,
                    "pool_args": {"n_sos": mol.n_active_sos}}
        adapt_solver = ADAPTSolver(opt_dict)
        adapt_solver.build()
        adapt_solver.simulate()

        self.assertAlmostEqual(adapt_solver.optimal_energy, -1.8945, places=3)

        deflation_circuits = [adapt_solver.optimal_circuit]
        ref_state = [0, 1, 0, 0, 0, 1]
        opt_dict = {"molecule": mol, "max_cycles": 1, "verbose": False, "pool": get_majorana_uccgsd_pool,
                    "pool_args": {"n_sos": mol.n_active_sos},
                    "deflation_circuits": deflation_circuits, "ref_state": ref_state, "deflation_coeff": 1}
        adapt_solver = ADAPTSolver(opt_dict)
        adapt_solver.build()
        optimal_energy = adapt_solver.simulate()

        self.assertAlmostEqual(optimal_energy, -1.91062, places=3)


if __name__ == "__main__":
    unittest.main()



"""
    Python wrappers around IonQ REST API, to facilitate job submission, result retrieval and post-processing
    Using IonQ services requires an API key.
    Users are expected to set the environment variable IONQ_APIKEY with the value of this token.
    IonQ documentation for the API: https://docs.ionq.co
    IonQ cloud dashboard: https://cloud.ionq.com/
"""

import os
import json
import time
from datetime import datetime
import pprint
import pandas as pd
import requests as rq

from tangelo.linq.translator import translate_json_ionq
from tangelo.linq.qpu_connection.qpu_connection import QpuConnection


class IonQConnection(QpuConnection):
    """ Wrapper about the IonQ REST API, to facilitate job submission and automated post-processing of results """

    def __init__(self):
        self.endpoint = "https://api.ionq.co" + "/v0.2"  # Update endpoint or version number here if needed
        self.api_key = None
        self._login()

    @property
    def header(self):
        """ Produce the header for REST requests """
        return {"Content-Type": "application/json", "Authorization": self.api_key}

    def _login(self):
        """ Retrieve the API key to be used for request with IonQ's REST API.
            Assumes users have the environment variable IONQ_APIKEY set to the correct value.
        """
        api_key = os.getenv("IONQ_APIKEY", None)
        if not api_key:
            raise RuntimeError(f"Please set these environment variables: IONQ_APIKEY")
        self.api_key = f'apiKey {api_key}'

        # Verify API key by submitting a trivial request
        try:
            self.job_get_history()
        except RuntimeError as err:
            raise RuntimeError(f"{err}")

    def _catch_request_error(self, return_dict):
        """ Use the dictionary returned from a REST request to check for errors at runtime, and catch them """
        if "error" in return_dict:
            pprint.pprint(return_dict)
            raise RuntimeError(f"Error returned by IonQ API :\n{return_dict['error']}")

    def _get_job_dataframe(self, job_history):
        """ Display main job info as pandas dataframe. Takes REST request answer as input

        Args:
            job_history (dict): job history REST output

        Returns:
            pandas.Dataframe: relevant info in dataframe format.
        """

        jl = job_history['jobs']
        jl_info = [(j['id'], j['status'], j['target']) for j in jl]
        jobs_df = pd.DataFrame(jl_info, columns=['id', 'status', 'target'])
        return jobs_df

    def _get_backend_dataframe(self, backend_info):
        """ Display backend info as pandas dataframe. Takes REST request answer as input

        Args:
            backend_info (dict): backend info REST output

        Returns:
            pandas.Dataframe: relevant info in dataframe format.
        """

        b_info = [(b['backend'], b['qubits'], b['status'],
                   datetime.utcfromtimestamp(b['last_updated']).strftime('%Y-%m-%d %H:%M:%S'),
                   b['average_queue_time'] // 10**6,
                   b.get('characterization_url', None))
                  for b in backend_info]

        df = pd.DataFrame(b_info, columns=['backend', 'qubits', 'status',
                                           'last updated', 'average queue time',
                                           'characterization_url'])
        return df

    def get_characterization(self, backend_name=None, charac_url=None):
        """ Retrieve characterization of a target device, either using the IonQ backend string,
        or an already-retrieved characterization url. The result contains information about
        number of qubits, fidelity of gates, t1, t2 and other specs, connectivity...

        Args:
            backend_name (str): string identifier for the IonQ backend
            charac_url (str): characterization url previously retrieved from IonQ API

        Returns:
            dict : a dictionary containing various information about the device
        """

        if bool(backend_name) == bool(charac_url):
            raise ValueError("Exactly one of these arguments need to be provided: backend_name, charac_url")
        if backend_name:
            endpoint = f"{self.endpoint}/characterizations/backends/{backend_name}/current"
        elif charac_url:
            endpoint = f"{self.endpoint}{charac_url}"

        job_request = rq.get(endpoint, headers=self.header)
        return_dict = json.loads(job_request.text)
        self._catch_request_error(return_dict)

        res = {'qubits': return_dict['qubits'], 'fidelity': return_dict['fidelity'],
               'timing': return_dict['timing'], 'connectivity': return_dict['connectivity']}
        return res

    def get_backend_info(self):
        """ Retrieve all the information available about the backends """

        job_request = rq.get(self.endpoint + '/backends', headers=self.header)
        return_dict = json.loads(job_request.text)

        self._catch_request_error(return_dict)
        return self._get_backend_dataframe(return_dict)

    def job_submit(self, target_backend, abs_circuit, n_shots, job_name, **job_specs):
        """ Submit job, return job ID.

        Args:
            target_backend (str): name of target device. See IonQ documentation for possible values.
                Current acceptable values are 'simulator' and 'qpu'
            abs_circuit (Circuit): Circuit in Tangelo format
            n_shots (int): number of shots (ignored if target_backend is set to `simulator`
            job_name (str): name to make the job more identifiable
            **job_specs: extra arguments such as `lang` in the code below; `metadata` is not currently supported.

        Returns:
            str: string representing the job id
        """

        # Convert abstract circuit into IonQ JSON format
        ionq_circuit = translate_json_ionq(abs_circuit)

        payload = {"target": target_backend,
                   "name": job_name,
                   "shots": n_shots,
                   "body": ionq_circuit,
                   "lang": job_specs.get('lang', 'json')
                   }

        job_request = rq.post(self.endpoint + '/jobs', headers=self.header, data=json.dumps(payload))
        return_dict = json.loads(job_request.text)

        self._catch_request_error(return_dict)
        print(f"Job submission \tID :: {return_dict['id']} \t status :: {return_dict['status']}")
        return return_dict['id']

    def job_get_history(self):
        """ Returns information about the job corresponding to the input job id

        Args:
            job_id (str): alphanumeric character string representing the job id

        Returns:
            dict: status response from the REST API
        """

        job_history = rq.get(self.endpoint + '/jobs', headers=self.header)
        return_dict = json.loads(job_history.text)

        self._catch_request_error(return_dict)
        return self._get_job_dataframe(return_dict)

    def job_status(self, job_id):
        """ Returns information about the job corresponding to the input job id

        Args:
            job_id (str): string representing the job id

        Returns:
            dict: status response from the REST API
        """

        job_status = rq.get(self.endpoint + "/jobs/" + job_id, headers=self.header)
        job_status = json.loads(job_status.text)
        self._catch_request_error(job_status)
        print(f"Job info \tID:: {job_id} \t status :: {job_status['status']} {job_status.get('error', '')}")
        return job_status

    def job_results(self, job_id):
        """ Blocking call querying the REST API at a given frequency, until job results are available.

        Args:
            job_id (str): string representing the job id

        Returns:
            dict: status response from the REST API
        """

        while True:
            job_status = self.job_status(job_id)
            if job_status['status'] == 'completed' and 'data' in job_status:
                hist = job_status['data']['histogram']
                h = dict()
                for k, v in hist.items():
                    bs = bin(int(k)).split('b')[-1]
                    h[("0"*(job_status['qubits']-len(bs)) + bs)[::-1]] = v
                return h
            elif job_status['status'] in {'ready', 'running', 'submitted'}:
                time.sleep(5)
            else:
                raise RuntimeError(f'Unexpected job status :: \n {job_status}')

    def job_cancel(self, job_id):
        """ Cancel / delete a job from IonQ servers.

        Args:
            job_id (str): string representing the job id

        Returns:
            dict: status response from the REST API
        """
        job_cancel = rq.delete(self.endpoint+"/jobs/"+job_id, headers=self.header)
        job_cancel = json.loads(job_cancel.text)
        self._catch_request_error(job_cancel)
        print(f"Job cancel \tID :: {job_id} \t status :: {job_cancel['status']} {job_cancel.get('error', '')}")
        return job_cancel
